%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% transformation of Flat-Curry programs into Prolog programs
%
% main predicate: (initializeCompilerState,) c2p(<progname>)

% required: environment variable PAKCSHOME

:- use_module(basics).
:- use_module(version).
:- use_module(library(system)).
:- use_module(external). % specification of external functions
:- use_module(readFlcFromFcy).

:- dynamic numberOfShares/1,
	   maxTupleArity/1, includePrelude/0,
	   bugInFlcFile/0,
	   allFunctions/1, allConstructors/1, externalFuncs/1, currentFunction/1,
	   newFunctionCounter/2, newAuxFunctions/1,
	   dynamicPredNames/1, loadProgInits/1.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Parameters for the transformation:

numberOfShares(0).   % count number of shared variables in a compiled program
includePrelude.      % uncomment this if prelude should not be included
                     % (in this case, only the main program will be translated
                     %  which is useful for compiler debugging)
maxTupleArity(7).    % The maximal arity of tuples

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

allModules([]).  % list of all modules (main + imported)
allFunctions([]).     % list of all defined functions
allConstructors([]).  % list of all constructors
externalFuncs([]).    % list of all external functions
currentFunction(unknown). % name of currently translated function (for errors)
newFunctionCounter("",0). % counter for generating new function names
newAuxFunctions([]). % list of generated auxiliary functions
                     % (for nested case/or)
dynamicPredNames([]). % list of pred names/file names for dynamic predicates
loadProgInits(true).  % Initialization directives to be performed
                      % after loading the program

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

addLoadProgInit(InitCmd) :-
	retract(loadProgInits(Inits)),
	asserta(loadProgInits((InitCmd,Inits))).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% for profiling:

% get cost center from function name:
getCostCenterOfName(NameS,CC) :-
	append(_,[123|NCC],NameS), % 123={
	append(CCS,[125|_],NCC), % 125=}
	append([123|CCS],[125],CCD),
	atom_chars(CC,CCD), !.
getCostCenterOfName(_,'').

deleteCostCenterInName(NameCC,Name) :-
	atom_chars(NameCC,NameCCS),
	append(Name0,[123|NCC],NameCCS), % 123={
	append(_,[125|Name1],NCC), % 125=}
	append(Name0,Name1,NameS),
	atom_chars(Name,NameS), !.
deleteCostCenterInName(Name,Name).

% get all cost centers from list of defined functions:
addCostCenterOfFuncs(CCs,[],CCs).
addCostCenterOfFuncs(CCs,['Func'(Name,_,_,_,_)|Funcs],NewCCs) :-
	getCostCenterOfName(Name,CC),
	(member(CC,CCs) -> CCs1=CCs ; append(CCs,[CC],CCs1)),
	addCostCenterOfFuncs(CCs1,Funcs,NewCCs).

% write a cost center name without decoration:
writeCostCenter(Stream,CC) :-
	atom_chars(CC,CCS),
	append([123|CCNS],[125],CCS),
	atom_chars(CCN,CCNS),
	write(Stream,' '), write(Stream,CCN).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

reportLiftBug :-
	writeErr('> Probably, this is due to a compiler bug in transforming'),
	nlErr,
	writeErr('> do/let/where expressions. Suggested solution:'),
	nlErr,
	writeErr('> Simplify dependencies in the do/let/where expression.'),
	nlErr,
	setFlcBug.

pleaseReport :-
	writeErr('*** Please report this error to the PAKCS developers'),
	nlErr,
	writeErr('*** if you have used the standard compiler!'),
	nlErr,
	setFlcBug.


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

c2p(Prog) :-
	appendAtom(Prog,'.pl',PrologFile),
	c2p(Prog,PrologFile).

c2p(Prog,PrologFile) :-
	(compileWithCompact(no) -> true ; compactFlatCurry(Prog)),
	split2dirbase(Prog,ProgDir,ModName),
	loadPath(ProgDir,LoadPath),
	(includePrelude
	 -> readProg(LoadPath,ModName,FlatProg,_),
	    'Prog'(MainModS,ImportStrings,_,_,_) = FlatProg,
	    map2(atom_chars,Imports,ImportStrings),
	    atom_chars(MainMod,MainModS),
	    readImportedEntities(LoadPath,Imports,[MainMod],
				 [],[],[],ImpTypes,ImpFuncs,ImpOps)
	  ; readProg(LoadPath,ModName,FlatProg,_),
	    ImpTypes=[], ImpFuncs=[], ImpOps=[]),
	generateProg(FlatProg,ImpTypes,ImpFuncs,ImpOps,PrologFile), !.
c2p(Prog,PrologFile) :-
	writeErr('*** Error during compilation of program "'),
	writeErr(Prog),
	writeErr('"!'), nlErr,
	appendAtom('rm -f ',PrologFile,RmCmd),
	system(RmCmd).

% compact a FlatCurry file by eliminating all unused functions:
compactFlatCurry(ProgAtom) :-
	standardLoadPath(LP),
	(LP=[] -> SetLP=""
	  ; path2String(LP,LPS),
	    append("CURRYPATH=",LPS,SLP1),
	    append(SLP1,"; export CURRYPATH; ",SetLP)),
	environ('PAKCSHOME',PH),
        atom_chars(PH,PHL),
	append(SetLP,PHL,PC0),
	append(PC0,"/tools/optimize/CompactFlat.state ",PC1),
	compileWithCompact(CompParams),
	append(PC1,CompParams,PC2),
	atom_chars(ProgAtom,ProgString),
	append(PC2,ProgString,PC3),
	append(PC3," >&2",PC4),
        atom_chars(CompactCmd,PC4),
        %write('Executing: '), write(CompactCmd), nl,
        system(CompactCmd),
	!.

% read all imported entities (types/functions/operators) of a set of modules:
readImportedEntities(_,[],_,ImpTypes,ImpFuncs,ImpOps,ImpTypes,ImpFuncs,ImpOps) :- !.
readImportedEntities(LoadPath,[Imp|Imps],ProcessedImps,
		     ImpTypes,ImpFuncs,ImpOps,AllImpTypes,AllImpFuncs,AllImpOps) :-
	member(Imp,ProcessedImps),
	!,
	%write('*** already read: '), write(Imp), nl,
	readImportedEntities(LoadPath,Imps,ProcessedImps,
			     ImpTypes,ImpFuncs,ImpOps,AllImpTypes,AllImpFuncs,AllImpOps).
readImportedEntities(LoadPath,[Imp|Imps],ProcessedImps,
		     ImpTypes,ImpFuncs,ImpOps,AllImpTypes,AllImpFuncs,AllImpOps) :-
	%write('*** read next: '), write(Imp), nl,
	(separateCompilation
	 -> readInterface(LoadPath,Imp,ImpInterface,AbsFlatIntFile)
	  ; readProg(LoadPath,Imp,ImpInterface,AbsFlatIntFile)),
	ImpInterface = 'Prog'(_,ImportStrings,Types,Funcs,Ops),
	map2(atom_chars,NewImports,ImportStrings),
	union(Imps,NewImports,AllImports),
	append(ImpTypes,Types,NewImpTypes),
	append(ImpFuncs,Funcs,NewImpFuncs),
	append(ImpOps,Ops,NewImpOps),
	readImportedEntities(LoadPath,AllImports,[Imp|ProcessedImps],
			     NewImpTypes,NewImpFuncs,NewImpOps,
			     AllImpTypes,AllImpFuncs,AllImpOps),
	getPrologFileNameFromFlat(AbsFlatIntFile,PrologFile),
	(doesPrologTranslationExists(AbsFlatIntFile,PrologFile) -> true
	 ; readProg(LoadPath,Imp,ImpProg,AbsFlatProgFile),
	   writeErrNQ('Compiling \''),
	   writeErrNQ(AbsFlatProgFile),
	   writeErrNQ('\' into \''),
	   writeErrNQ(PrologFile),
	   writeErrNQ('\'...'),
	   generateProg(ImpProg,AllImpTypes,AllImpFuncs,AllImpOps,PrologFile),
	   writeErrNQ('done'), nlErrNQ).

getPrologFileNameFromFlat(AbsFlatProgFile,PrologFile) :-
	atom_chars(AbsFlatProgFile,AbsFlatProgFileS),
	(append(AbsProgFileS,".fcy",AbsFlatProgFileS) -> true
	 ; append(AbsProgFileS,".fint",AbsFlatProgFileS)),
	append(AbsProgFileS,".pl",AbsProgPlFileS),
	atom_chars(PrologFile,AbsProgPlFileS).

doesPrologTranslationExists(_,_) :-
	\+ separateCompilation, % translated Prolog files only required for sep. comp.
	!.
doesPrologTranslationExists(AbsFlatProgFile,PrologFile) :-
	checkProgramHeader(PrologFile),
	fileExistsAndNewer(PrologFile,AbsFlatProgFile),
        %writeErrNQ('Compiled Curry program \''),
	%writeErrNQ(PrologFile),
	%writeErrNQ('\' is up-to-date.'), nlErrNQ,
	!.


% initialize the state of the compiler, i.e., clean all tables and
% set all parameters to their default values:
initializeCompilerState :-
	%(retract(includePrelude) -> true ; true),
	%asserta(includePrelude),
	(retract(numberOfShares(_)) -> true ; true),
	asserta(numberOfShares(0)),
	(retract(allFunctions(_)) -> true ; true),
	asserta(allFunctions([])),
	(retract(allConstructors(_)) -> true ; true),
	asserta(allConstructors([])),
	(retract(externalFuncs(_)) -> true ; true),
	asserta(externalFuncs([])),
	(retract(newAuxFunctions(_)) -> true ; true),
	asserta(newAuxFunctions([])),
	(retract(dynamicPredNames(_)) -> true ; true),
	asserta(dynamicPredNames([])),
	(retract(loadProgInits(_)) -> true ; true),
	asserta(loadProgInits(true)),
	% remove all non-standard infix declarations to avoid syntax errors
	% in compiled program:
	op(0,xfx,(+#)),
	op(0,xfx,(-#)),
	op(0,xfx,(*#)),
	op(0,xfx,(/#)),
	op(0,xfx,(=#)),
	op(0,xfx,(/=#)),
	op(0,xfx,(<#)),
	op(0,xfx,(<=#)),
	op(0,xfx,(>#)),
	op(0,xfx,(>=#)),
	op(0,xfx,(+.)),
	op(0,xfx,(-.)),
	op(0,xfx,(*.)),
	op(0,xfx,(/.)),
	op(0,xfx,(<.)),
	op(0,xfx,(<=.)),
	op(0,xfx,(>.)),
	op(0,xfx,(>=.)).


generateProg('Prog'(Mod,Imports,MainTypes,MainFuncs,MainOps),
	            ImpTypes,ImpFuncs,ImpOps,PrologFile) :-
	initializeCompilerState,
	(compileWithDebug
	 -> writeErr('...including code for debugging'),
	    nlErr
	  ; true),
	(compileWithFailPrint
	 -> writeErr('...including code for failure printing'),
	    nlErr
	  ; true),
	tryWriteFile(PrologFile),
	tell(PrologFile),
	writePrologHeader,
	writeClause((:- prolog_flag(single_var_warnings,_,off))),
	writeClause((:- prolog_flag(redefine_warnings,_,off))),
	nl,
	(separateCompilation -> map1(writeLoadImport,Imports), nl ; true),
	writeProg(Mod,Imports,MainTypes,MainFuncs,MainOps,
		  ImpTypes,ImpFuncs,ImpOps),
	nl,
	loadProgInits(ProgInits),
	(ProgInits=true -> true
	                 ; writeClause((:- initializationsInModule(ProgInits))), nl),
	dynamicPredNames(DynPreds),
	map1(writeDynamicInfoClause,DynPreds),
	nl,
	write('%%%%% Number of shared variables: '),
	numberOfShares(SC), write(SC), nl,
	told, !.
generateProg(_,PrologFile) :-
	told,
	writeErr('*** Error during compiling, no program generated!'),
	nlErr,
	appendAtom('rm -f ',PrologFile,RmCmd),
	system(RmCmd).

% write head of generated Prolog file:
writePrologHeader :-
	write('%'),
	compilerVersion(CV),
	write(CV),
	(separateCompilation -> write(' SEPARATE')
	 ; (compileWithDebug -> write(' DEBUG') ; true),
	   (compileWithFailPrint -> write(' FAILPRINT') ; true)),
	(compileWithSharing(variable) -> write(' VARIABLESHARING') ; true),
	(compileWithSharing(function) -> write(' FUNCTIONSHARING') ; true),
	nl, nl.

writeLoadImport(ImpModS) :-
	atom_chars(ImpMod,ImpModS),
	writeClause((:- importModule(ImpMod))).

% compute all external libraries to be included for externally defined functions:
getExternalLibraries([],Libs,Libs) :- !.
getExternalLibraries(['Func'(_,_,_,_,'External'(EStr))|Funcs],Libs,AllLibs) :-
	append(LibS,[32|_],EStr), % ext.names have the form "lib ename"
	!,
	atom_chars(Lib,LibS),
	(member(Lib,Libs) -> Libs1=Libs ; Libs1=[Lib|Libs]),
	getExternalLibraries(Funcs,Libs1,AllLibs).
getExternalLibraries([_|Funcs],Libs,AllLibs) :-
	getExternalLibraries(Funcs,Libs,AllLibs).

% check for necessary inclusion of standard libraries (ports, CLP,...):
writeLibraryInclusions(Libs) :-
	(member(prim_dynamic,Libs)
	 -> addLoadProgInit(initializeDynamic)
	  ; true),
	(member(prim_ports,Libs)
	 -> % if Ports are used, ensure to start CPNS demon
	    addLoadProgInit(start_cpns_demon)
	  ; true),
	map1(writeLibraryInclusion,Libs).

writeLibraryInclusion(Lib) :-
	writeClause((:- ensure_lib_loaded(Lib))).


% try translating XML file into .fcy file:
tryXml2Fcy(FlatFile) :-
	atom_chars(FlatFile,FlatFileS),
	append(ProgS,".fcy",FlatFileS),
	append(ProgS,".curry",CurryFileS),
	atom_chars(CurryFile,CurryFileS),
	append(ProgS,"_flat.xml",XmlFileS),
	atom_chars(XmlFile,XmlFileS),
	\+ file_exists(CurryFile),
	\+ file_exists(FlatFile),
	file_exists(XmlFile),
	% translate XML file into .fcy file:
	environ('PAKCSHOME',PH),
	atom_chars(PH,PHL),
	append(PHL,"/tools/curry2xml.state -fcy ",CP1),
	append(CP1,ProgS,CP2),
	atom_chars(Cmd,CP2),
	%write('Executing: '), write(Cmd), nl,
	system(Cmd),
	!.
tryXml2Fcy(_).


% read a FlatCurry interface:
readInterface(LoadPath,Prog,FlatProg,AbsFlatProgFile) :-
	appendAtom(Prog,'.fint',ProgFcy),
	readInterfaceInLoadPath(LoadPath,ProgFcy,FlatProg,AbsFlatProgFile).

readInterfaceInLoadPath([],FlatFile,_,_) :-
	write('ERROR: FlatCurry file '), write(FlatFile), write(' not found!'), nl,
	!, fail.
readInterfaceInLoadPath([Dir|Dirs],ProgFile,FlatInt,AbsFlatIntFile) :-
	appendAtom(Dir,'/',DirSlash),
	appendAtom(DirSlash,ProgFile,DirProgFile),
	(file_exists(DirProgFile)
	 -> readFlcFromFcy(DirProgFile,FlatInt),
	    AbsFlatIntFile = DirProgFile,
	    (pakcsrc(showfcyload,yes)
	     -> checkForFurtherFcyProgs(Dir,Dirs,ProgFile) ; true)
	  ; readInterfaceInLoadPath(Dirs,ProgFile,FlatInt,AbsFlatIntFile)).

% read a flat program:
readProg(LoadPath,Prog,FlatProg,AbsFlatProgFile) :-
	appendAtom(Prog,'.fcy',ProgFcy),
	readProgInLoadPath(LoadPath,ProgFcy,FlatProg,AbsFlatProgFile).

readProgInLoadPath([],FlatFile,_,_) :-
	write('ERROR: FlatCurry file '), write(FlatFile), write(' not found!'), nl,
	!, fail.
readProgInLoadPath([Dir|Dirs],ProgFile,FlatProg,AbsFlatProgFile) :-
	appendAtom(Dir,'/',DirSlash),
	appendAtom(DirSlash,ProgFile,DirProgFile),
	tryXml2Fcy(DirProgFile),
	(file_exists(DirProgFile)
	 -> readFlcFromFcy(DirProgFile,PlainFlatProg),
	    AbsFlatProgFile = DirProgFile,
	    mergeWithPrimitiveSpecs(PlainFlatProg,DirProgFile,FlatProg),
	    (pakcsrc(showfcyload,yes)
	     -> checkForFurtherFcyProgs(Dir,Dirs,ProgFile) ; true)
	  ; readProgInLoadPath(Dirs,ProgFile,FlatProg,AbsFlatProgFile)).

checkForFurtherFcyProgs(_,[],_).
checkForFurtherFcyProgs(FcyDir,[Dir|Dirs],ProgFile) :-
	appendAtom(Dir,'/',DirSlash),
	appendAtom(DirSlash,ProgFile,DirProgFile),
	((file_exists(DirProgFile), \+ equalDirectories(FcyDir,Dir))
	 -> writeErrNQ('WARNING: further FlatCurry file found (but ignored): '),
	    writeErrNQ(DirProgFile), nlErrNQ,
	    checkForFurtherFcyProgs(FcyDir,Dirs,ProgFile)
	  ; checkForFurtherFcyProgs(FcyDir,Dirs,ProgFile)).

% are two directories identical?
equalDirectories(Dir1,Dir2) :-
	% a bit complicate solution, but this works also for links
	absolute_file_name(Dir1,AbsDir1),
	absolute_file_name(Dir2,AbsDir2),
	working_directory(WDir,AbsDir1), absolute_file_name('.',AbsDir1a),
	working_directory(_   ,AbsDir2), absolute_file_name('.',AbsDir2a),
	working_directory(_,WDir),
	!,
	AbsDir1a=AbsDir2a.


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% try to read primitive function specification file and merge infos into external
% functions:
mergeWithPrimitiveSpecs(PlainFlatProg,DirProgFcy,FlatProg) :-
	PlainFlatProg = 'Prog'(ModName,_,_,_,_),
	atom_chars(DirProgFcy,DirProgFcyS),
	append(DirProgS,".fcy",DirProgFcyS),
	append(DirProgS,".prim_c2p",DirProgExtS),
	atom_chars(PrimXmlFile,DirProgExtS),
	file_exists(PrimXmlFile),
	!,
	(pakcsrc(showfcyload,yes)
	 -> writeErr('>>> Reading '),
	    writeErr(PrimXmlFile), writeErr(' ... '),
	    statistics(runtime,[RT1,_]) ; true),
	readPrimitiveXmlSpecs(PrimXmlFile,PrimSpecs),
	(pakcsrc(showfcyload,yes)
	 -> statistics(runtime,[RT2,_]),
	    RT is RT2-RT1,
	    writeErr(RT), writeErr(' ms.'), nlErr
	  ; true),
        append(ModName,".",ModNameDotS), atom_chars(ModNameDot,ModNameDotS),
	map2partial(addModuleName2PrimSpecs(ModNameDot),PrimSpecs,QPrimSpecs),
	addPrimitiveSpecs2FlatProg(PlainFlatProg,QPrimSpecs,FlatProg).
mergeWithPrimitiveSpecs(PlainFlatProg,_,FlatProg) :-
	addPrimitiveSpecs2FlatProg(PlainFlatProg,[],FlatProg).

addModuleName2PrimSpecs(ModNameDot,primitive(F,N,Mod,Entry),primitive(F,N,Mod,Entry)) :-
	ModNameDot='prelude.', !.
addModuleName2PrimSpecs(ModNameDot,primitive(F,N,Mod,Entry),primitive(QF,N,Mod,Entry)) :-
	appendAtom(ModNameDot,F,QF).
addModuleName2PrimSpecs(ModNameDot,ignore(F,N),ignore(F,N)) :- ModNameDot='prelude.', !.
addModuleName2PrimSpecs(ModNameDot,ignore(F,N),ignore(QF,N)) :-
	appendAtom(ModNameDot,F,QF).

addPrimitiveSpecs2FlatProg('Prog'(ModName,Imps,Types,Funcs,Ops),PrimSpecs,
                           'Prog'(ModName,Imps,Types,ModFuncs,Ops)) :-
        addPrimitiveSpecs2Funcs(PrimSpecs,Funcs,ModFuncs).

% format a primitive specification (e.g., for error messages):
writePrimSpec(primitive(F,Arity,ELib,_EName)) :-
	writeErr(F), writeErr('/'), writeErr(Arity), writeErr(' in library "'),
	writeErr(ELib), writeErr('"'), nlErr.
writePrimSpec(ignore(F,Arity)) :-
	writeErr(F), writeErr('/'), writeErr(Arity), nlErr.

% process list of function declarations w.r.t. primitive specification:
addPrimitiveSpecs2Funcs([],[],[]) :- !.
addPrimitiveSpecs2Funcs(PrimSpecs,[],[]) :-
	writeErrNQ('WARNING: specifications of primitive functions '),
	writeErrNQ('without source code found:'), nlErrNQ,
	(quietmode(no) -> map1(writePrimSpec,PrimSpecs) ; true), !.
addPrimitiveSpecs2Funcs(PrimSpecs,['Func'(Name,_,_,Type,_)|Funcs],MFuncs) :-
	argnum(Type,Arity), flatName2Atom(Name,F), deleteCostCenterInName(F,FWOCC),
	deleteFirst(ignore(FWOCC,Arity),PrimSpecs,DPrimSpecs), !,
	addPrimitiveSpecs2Funcs(DPrimSpecs,Funcs,MFuncs).
addPrimitiveSpecs2Funcs(PrimSpecs,
			['Func'(Name,_,Vis,Type,_)|Funcs],
     	                ['Func'(Name,Arity,Vis,Type,'External'(EStr))|MFuncs]) :-
	argnum(Type,Arity), flatName2Atom(Name,F), deleteCostCenterInName(F,FWOCC),
	deleteFirst(primitive(FWOCC,Arity,EMod,EName),PrimSpecs,DPrimSpecs), !,
	atom_chars(EMod,EModS),
	atom_chars(EName,ENameS),
	append(EModS,[32|ENameS],EStr),
	addPrimitiveSpecs2Funcs(DPrimSpecs,Funcs,MFuncs).
addPrimitiveSpecs2Funcs(PrimSpecs,
			['Func'(Name,Arity,Vis,Type,'External'(EStr))|Funcs],
			['Func'(Name,Arity,Vis,Type,'External'(EStr))|MFuncs]) :-
	append(_,[32|_],EStr), % has the form of an already known primitive
	!,
	addPrimitiveSpecs2Funcs(PrimSpecs,Funcs,MFuncs).
addPrimitiveSpecs2Funcs(_,['Func'(Name,_,_,Type,'External'(_))|_],_) :-
	argnum(Type,Arity), flatName2Atom(Name,F), deleteCostCenterInName(F,FWOCC),
	writeErr('ERROR: specification of primitive function '),
	writeErr(FWOCC), writeErr('/'), writeErr(Arity),
	writeErr(' not found!'), nlErr, !, setFlcBug, fail.
addPrimitiveSpecs2Funcs(PrimSpecs,[Func|Funcs],[Func|MFuncs]) :-
	addPrimitiveSpecs2Funcs(PrimSpecs,Funcs,MFuncs).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% add the predefined objects not explicitly contained in the FlatPrelude:
addPredefinedObjects('Prog'(ModName,Imps,Types,Funcs,Ops),
                     'Prog'(ModName,Imps,Types,Funcs,Ops)) :-
        % predefined types already contained?
        member('Type'("Int",'Public',[],[]),Types),
        !.
addPredefinedObjects('Prog'(ModName,Imps,Types,Funcs,Ops),
                     'Prog'(ModName,NImps,NTypes,Funcs,Ops)) :-
        % delete "prelude" from import list due to bug in flc printer:
        writeErrNQ('WARNING: Adding predefined objects!'), nlErrNQ,
        deleteFirst("prelude",Imps,NImps),
        NTypes = [
               'Type'("Int",'Public',[],[]),
               'Type'("Float",'Public',[],[]),
               'Type'("Char",'Public',[],[]),
               'Type'("Bool",'Public',[],['Cons'("True",0,'Public',[]),
					  'Cons'("False",0,'Public',[])]),
               'Type'("(,)",'Public',[0,1],['Cons'("(,)",2,'Public',
						   ['TVar'(0),'TVar'(1)])]),
               'Type'("(,,)",'Public',[0,1,2],
		      ['Cons'("(,,)",3,'Public',['TVar'(0),'TVar'(1),'TVar'(2)])]),
               'Type'("(,,,)",'Public',[0,1,2,3],
		      ['Cons'("(,,,)",4,'Public',
			      ['TVar'(0),'TVar'(1),'TVar'(2),'TVar'(3)])]),
               'Type'("(,,,,)",'Public',[0,1,2,3,4],
		      ['Cons'("(,,,,)",5,'Public',
			      ['TVar'(0),'TVar'(1),'TVar'(2),
			       'TVar'(3),'TVar'(4)])]),
               'Type'("(,,,,,)",'Public',[0,1,2,3,4,5],
		      ['Cons'("(,,,,,)",6,'Public',
			      ['TVar'(0),'TVar'(1),'TVar'(2),
			       'TVar'(3),'TVar'(4),'TVar'(5)])]),
               'Type'("(,,,,,,)",'Public',[0,1,2,3,4,5,6],
		      ['Cons'("(,,,,,,)",7,'Public',
			      ['TVar'(0),'TVar'(1),'TVar'(2),
			       'TVar'(3),'TVar'(4),'TVar'(5),'TVar'(6)])]),
               'Type'("[]",'Public',[0],['Cons'("[]",0,'Public',[]),
                                'Cons'(":",2,'Public',
				      ['TVar'(0),'TCons'("[]",['TVar'(0)])])]),
               'Type'("IO",'Public',[0],['Cons'("$io",1,'Public',['TVar'(0)])]),
               'Type'("()",'Public',[],['Cons'("()",0,'Public',[])]),
               'Type'("Success",'Public',[],['Cons'("success",0,'Public',[])]) | Types].


writeProg(Mod,Imports,MainTypes,MainFuncs,MainOps,ImpTypes,ImpFuncs,ImpOps) :-
	append(MainTypes,ImpTypes,AllTypes),
	(separateCompilation -> CodeTypes=MainTypes ; CodeTypes=AllTypes),
	(separateCompilation -> ImportedFuncs=ImpFuncs, CodeFuncs=MainFuncs
	                      ; ImportedFuncs=[], append(MainFuncs,ImpFuncs,CodeFuncs)),
	append(MainOps,ImpOps,AllOps),
	getExternalLibraries(CodeFuncs,[],ExtLibs),
	writeLibraryInclusions(ExtLibs), nl,
	atom_chars(ModName,Mod),
	writeClause((:- curryModule(ModName))), nl,
	% redundant checking, could be omitted if front end is correct:
	(retract(bugInFlcFile) -> true ; true),
	map1(check_flcFunction,CodeFuncs),
	\+ bugInFlcFile,
	(plprofiling(yes) -> addCostCenterOfFuncs([''],CodeFuncs,CCs) ; CCs=['']),
	CCs=[_|NewCCs],
	(NewCCs=[] -> true
	 ; writeErr('...including code for profiling cost centers:'),
	   map1partial(writeCostCenter(user_error),NewCCs),
	   nlErr),
	computeAllExternalFunctions(CodeFuncs,ExtFuncs),
	(retract(externalFuncs(_)) -> true ; true),
	asserta(externalFuncs(ExtFuncs)),
	(map2(elimNestedOrCases,CodeFuncs,CodeFuncsWOnestedcase)
	 -> true
	  ; writeErr('INTERNAL COMPILER ERROR in or/case lifter!'), nlErr, fail),
	newAuxFunctions(TmpAuxFuns), rev(TmpAuxFuns,RevAuxFuns),
	append(CodeFuncsWOnestedcase,AllTypes,CodeFuncsAndAllTypes),
	(separateCompilation -> append(CodeFuncsAndAllTypes,ImportedFuncs,AllFunData)
			      ; AllFunData = CodeFuncsAndAllTypes),
	computeCorrectType(AllFunData,RevAuxFuns,AuxFuns),
	append(CodeFuncsWOnestedcase,AuxFuns,CodeFuncsOIS),
	(map2(transformFunction2FirstOrder,CodeFuncsOIS,CodeFuncsOIS1st)
	 -> true
	  ; writeErr('INTERNAL COMPILER ERROR in higher-order elimination!'), nlErr,
	    fail),
	append(CodeFuncsOIS1st,ImportedFuncs,AllFuncs),
	map2(flcFunc2FA,AllFuncs,AllFuncsArities),
	retract(allFunctions(_)), asserta(allFunctions(AllFuncsArities)),
	write('%%%%%%%%%%%% function types %%%%%%%%%%%%%%%%%%%'), nl,
	(separateCompilation -> writeClause((:- multifile functiontype/6)) ; true),
	writeClause((:- dynamic functiontype/6)),
	map1partial(writeFTypeClause(ExtFuncs,AllOps),CodeFuncsOIS1st), nl,
	write('%%%%%%%%%%%% constructor types %%%%%%%%%%%%%%%%%%%'), nl,
	(separateCompilation -> writeClause((:- multifile constructortype/4)) ; true),
	writeClause((:- dynamic constructortype/4)),
	((separateCompilation, member("prelude",Imports)) -> true
	 ; % generate type clause for partcall only in the prelude
	   % in case of separate compilation:
	   writeClause(constructortype(partcall,partcall,3,
	    'FuncType'('TCons'('Int',[]),'FuncType'(_,'FuncType'('TCons'([],[_]),_)))))),
	map1(writeDTypeClause,CodeTypes), nl,

	getConstructors(AllTypes,ConsList),
	retract(allConstructors(_)), asserta(allConstructors(ConsList)),
	write('%%%%%%%%%%%% function definitions %%%%%%%%%%%%%%%%%%%'), nl,
	map1(writeFunc,CodeFuncsOIS1st),

	(separateCompilation
	 -> writeClause((:- costCenters(CCs))), nl
	  ; write('%%%%%%%%%%%% hnf clauses %%%%%%%%%%%%%%%%%%%'), nl,
	    trans_hnf(CodeFuncsOIS1st,ExtFuncs,CCs),
	    writeGenericClauses(CCs)),
	!,
	\+ bugInFlcFile,
	nl.

writeGenericClauses(CCs) :-
	write('%%%%%%%%%%%% clauses for generic operations %%%%%%%%%%%%%%%%%%%'), nl,
	map1(trans_constrEq,CCs),
	map1(trans_boolEq,CCs),
	map1(trans_nf,CCs),
	(compileWithSharing(variable) -> map1(transpropshar,CCs) ; true),
	(compileWithSharing(function) -> map1(genMakeFunctionShare,CCs) ; true),
	trans_deref(CCs).

% "Linker" for separate compilation:
% create a file containing hnf clauses for all currently loaded functions
% and the generic clauses to define =:=, ==, nf, etc:
generateMainPlFile(PrologFile,MainFile) :-
	currentCostCenters(CCs),
	appendAtom(PrologFile,'.main',MainFile),
	tell(MainFile),
	write('%%%%%%%%%%%% hnf clauses %%%%%%%%%%%%%%%%%%%'), nl,
	trans_hnf_current(CCs),
	writeGenericClauses(CCs),
	told.

% create main file, load it and delete it if not in profiling mode:
loadMain(PrologFile) :-
	separateCompilation, !,
	generateMainPlFile(PrologFile,MainPrologFile),
	compile(MainPrologFile),
	deleteMainPrologFile(MainPrologFile).
loadMain(_).

deleteMainPrologFile(MainPrologFile) :-
	(plprofiling(yes)
	 -> true % prog.pl.main needed later for PlProfileData.getHnfDefinitions
          ; % delete MainPrologFile:
	    appendAtom('rm -f ',MainPrologFile,RemoveMainPrologFile),
	    system(RemoveMainPrologFile)).
	
% a call to this predicate sets the global error flat so that no compiled
% program is produced:
setFlcBug :- bugInFlcFile -> true ; assertz(bugInFlcFile).

% check for deprecated functions and print warning:
checkForDeprecatedFunction(Name) :-
	member(Name,['Term.readTerm','Term.readQTerm','Term.showTerm',
		     'Term.showQTerm','ReadShowTerm.readTerm','ReadShowTerm.readsTerm',
		     'System.getDate','System.showDate',
		     'FlatTools.writeFLC','FlatCurryTools.writeFLC',
		     'HTML.Form','Ports.openSocketConnectPort']), !,
	writeErrNQ('WARNING: do not use deprecated function "'),
	writeErrNQ(Name),
	writeErrNQ('".'),
	nlErrNQ.
checkForDeprecatedFunction(_).

% check for ocurrences of constructors of Dynamic type and print error:
checkForDynamicConstructor(Name) :-
	member(Name,['Dynamic.dynamic','Dynamic.persistent']),
	currentFunction(QFunc),
	!,
	writeErr('ERROR in "'),
	writeErr(QFunc),
	writeErr('": Function "'),
	writeErr(Name),
	writeErr('" is not allowed in this context!'),
	nlErr,
	setFlcBug.
checkForDynamicConstructor(_).

% check for maximal arity of tuples and print warning:
checkForTupleArity([40,44|Name]) :-
	length([44|Name],TA),
	maxTupleArity(MA),
	TA>MA, !,
	currentFunction(QFunc),
	writeErr('ERROR in "'),
	writeErr(QFunc),
	writeErr('": arity of '), writeErr(TA),
	writeErr('-tupel too large. Reconfigure your system (see'), nlErr,
	writeErr('file "'), environ('PAKCSHOME',PH), writeErr(PH),
	writeErr('/parser/config.txt" for details).'),
	nlErr, setFlcBug.
checkForTupleArity(_).


% compute all external functions in this program:
computeAllExternalFunctions([],[]).
computeAllExternalFunctions(['Func'(Name,_,_,Type,'External'(EStr))|Funcs],
			    [(F/Arity,EStr)|EFs]) :- !,
	argnum(Type,Arity),
	flatName2Atom(Name,F),
	computeAllExternalFunctions(Funcs,EFs).
computeAllExternalFunctions([_|Funcs],EFs) :-
	computeAllExternalFunctions(Funcs,EFs).


% transform Flat-Curry representation of a function into name/arity pair
% where the arity is the arity of the final (eta-expanded) function:
flcFunc2FA('Func'(Name,_,_,Type,_),F/Arity) :- flatName2Atom(Name,F), argnum(Type,Arity).

% transform a FlatCurry name (string) into atom:
flatName2Atom(":",'.') :- !. % translate Curry list cons into Prolog list cons
flatName2Atom(".",'prelude..') :- !. % to avoid confusion, rename Curry's compose
flatName2Atom(Name,Atom) :- atom_chars(Atom,Name).
	

getConstructors([],Cs) :-
	TupleCons = ['()'/0,'(,)'/2,'(,,)'/3,'(,,,)'/4,'(,,,,)'/5],
	(includePrelude
          -> Cs=[partcall/3,'$stream'/1] % all standard constructors defined in prelude
           ; Cs=[partcall/3,'True'/0,'False'/0,success/0,[]/0,'.'/2|TupleCons]).
getConstructors(['Type'(_,_,_,DataCons)|Types],AllCons) :-
	getDataCons(DataCons,DCs),
	getConstructors(Types,TCs),
	append(DCs,TCs,AllCons).

getDataCons([],[]).
getDataCons(['Cons'(Name,Arity,_,_)|DataCons],[Con/Arity|DCs]) :-
	flatName2Atom(Name,Con),
	getDataCons(DataCons,DCs).

% compute the arity of a constructor from the list of all stored constructors
% (fail if arity not found):
getConsArity(Cons,Arity) :-
	allConstructors(CAs), getArity(CAs,Cons,Arity).

% compute the arity of a function from the list of all stored functions:
% (error message and fail if arity not found):
getFuncArity(Func,Arity) :-
	allFunctions(FAs), getArity(FAs,Func,Arity), !.
getFuncArity(Func,Arity) :- % special case for tuples:
	atom_chars(Func,[40,44|Commas]),
	length([44|Commas],Arity), !.
getFuncArity(Func,_) :-
	writeErr('INTERNAL COMPILER ERROR: Arity of function '),
	writeErr(Func),
	writeErr(' (used in function '), currentFunction(CF),writeErr(CF), writeErr(')'),
	writeErr(' not found!'), nlErr, !, fail.

getArity([Name/Arity|_],Name,Arity) :- !.
getArity([_|FAs],Name,Arity) :- getArity(FAs,Name,Arity).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% pre-transformation of defined functions:
% possibly add missing arguments in case of higher-order equations,
% i.e., in this implementation all functions are defined by first-order
% rules

% compute argument number (of uncurryfied function) from the functions type:
argnum('FuncType'(_,T2),N) :-
	argnum(T2,N2), N is N2+1.
argnum('TVar'(_),0).
argnum('TCons'(TC,_),N) :- TC="IO" -> N=1 ; N=0.

% compute maximum variable index in a term (list):
maxVarIndex('Var'(V),V).
maxVarIndex('Lit'(_),-1).
maxVarIndex('Comb'(_,_,Es),M) :- map2(maxVarIndex,Es,Ms), maxList(Ms,M).
maxVarIndex('Apply'(E1,E2),M) :-
	maxVarIndex(E1,M1), maxVarIndex(E2,M2), max(M1,M2,M).
maxVarIndex('Constr'(Vs,E),M) :-
	maxList(Vs,Ms), maxVarIndex(E,ME), max(Ms,ME,M).
maxVarIndex('Or'(E1,E2),M) :-
	maxVarIndex(E1,M1), maxVarIndex(E2,M2), max(M1,M2,M).
maxVarIndex('Case'(_,E,Cs),M) :-
	maxVarIndex(E,ME), map2(maxVarCase,Cs,Ms), maxList(Ms,MC),
	max(MC,ME,M).
maxVarIndex('GuardedExpr'(Vs,C,E),M) :-
	maxList(Vs,Ms), maxVarIndex(C,MC), maxVarIndex(E,ME),
	max(Ms,MC,MSC), max(MSC,ME,M).
maxVarIndex('Choice'(E),M) :- maxVarIndex(E,M).

maxVarCase('Branch'('Pattern'(_,Vs),E),M) :-
	maxList(Vs,Ms), maxVarIndex(E,ME), max(Ms,ME,M).
maxVarCase('Branch'('LPattern'(_),E),M) :-
	maxVarIndex(E,M).

% maximum of a list of naturals:
maxList([],-1).
maxList([X|Xs],M) :- maxList(Xs,Ms), max(X,Ms,M).

max(X,Y,Z) :- X>Y -> Z=X ; Z=Y.

% for simplicity, we generate negative variable indices here:
genVarIndices(_,0,[]).
genVarIndices(Max,N,[I|Is]) :-
	N>0, I is Max+1, N1 is N-1,
	genVarIndices(I,N1,Is).


% transform higher-order defining rules into first-order rules
% by eta-expansion:
transformFunction2FirstOrder('Func'(Name,Arity,Vis,Type,'Rule'(Args,RHS)),
		             'Func'(Name,NewArity,Vis,Type,'Rule'(NewArgs,NewRHS))) :-
	!,
	argnum(Type,NewArity),
	NNewArgs is NewArity-Arity,
	(NNewArgs<0
	  -> writeErrNQ('WARNING: Function "'),
	     flatName2Atom(Name,FName), writeErrNQ(FName),
	     writeErrNQ('" has illegal type!'), nlErrNQ
	   ; true),
	(NNewArgs=<0
	 -> NewArgs=Args, NewRHS=RHS % don't add any arguments
	  ; maxList(Args,MaxArgs), maxVarIndex(RHS,MaxRHS),
	    max(MaxArgs,MaxRHS,MaxIndex),
	    genVarIndices(MaxIndex,NNewArgs,AddArgs),
	    append(Args,AddArgs,NewArgs),
	    addArgsToRHS(RHS,AddArgs,NewRHS)).
	%freeVarsInExp(NewRHS,Vs),
	%diff(Vs,NewArgs,FVs),
	%(FVs=[] -> true ; writeErr(NewRHS),nlErr).
	%writeErr(NewRHS),nlErr,
	%writeErr(Vs),nlErr.
transformFunction2FirstOrder('Func'(Name,_Arity,Vis,Type,'External'(EName)),
			     'Func'(Name,NewFArity,Vis,Type,'External'(EName))) :- !,
	argnum(Type,NewFArity).
transformFunction2FirstOrder(Arg,Arg) :-
	writeErr('ERROR: Illegal argument in transformFunction2FirstOrder: '),
	writeErr(Arg),
	nlErr.

% add new variables to right-hand side, i.e., perform eta-expansion:
addArgsToRHS('Case'(CT,C,Cases),AddArgs,'Case'(CT,C,NewCases)) :- !,
	addArgsToBranches(Cases,AddArgs,NewCases).
addArgsToRHS('Or'(E1,E2),AddArgs,'Or'(NE1,NE2)) :- !,
	addArgsToRHS(E1,AddArgs,NE1),
	addArgsToRHS(E2,AddArgs,NE2).
addArgsToRHS('Choice'(E),AddArgs,'Choice'(NE)) :- !,
	addArgsToRHS(E,AddArgs,NE).
addArgsToRHS('GuardedExpr'(Vs,Con,Exp),AddArgs,'GuardedExpr'(Vs,Con,NExp)) :-
	!,
	addArgsToRHS(Exp,AddArgs,NExp).
addArgsToRHS(Exp,[],Exp) :- !.
addArgsToRHS(Exp,[A|As],NewExp) :- !,
	addArgsToRHS('Apply'(Exp,'Var'(A)),As,NewExp).

addArgsToBranches([],_,[]).
addArgsToBranches(['Branch'(Pat,Exp)|Cs],AddArgs,
	       ['Branch'(Pat,NewExp)|NCs]) :-
	addArgsToRHS(Exp,AddArgs,NewExp),
	addArgsToBranches(Cs,AddArgs,NCs).


% eliminate nested or/case expressions in right-hand sides of function
% definitions:
elimNestedOrCases('Func'(Name,Arity,Vis,Type,'Rule'(Args,RHS)),
		  'Func'(Name,Arity,Vis,Type,'Rule'(Args,NewRHS))) :-
	!,
	(retract(newFunctionCounter(_,_)) -> true ; true),
	asserta(newFunctionCounter(Name,0)),
	elimNestedCasesInRHS(RHS,NewRHS).
elimNestedOrCases('Func'(Name,Arity,Vis,Type,'External'(EName)),
		  'Func'(Name,Arity,Vis,Type,'External'(EName))) :- !.
elimNestedOrCases(Arg,Arg) :-
	writeErr('ERROR: Illegal argument in elimNestedOrCases: '),
	writeErr(Arg),
	nlErr.

% remove nested cases in right-hand side:
elimNestedCasesInRHS('Case'(CT,CE,Cases),'Case'(CT,NCE,NewCases)) :- !,
	elimCasesInExp(CE,NCE),
	elimNestedCasesInBranches(Cases,NewCases).
elimNestedCasesInRHS('Or'(E1,E2),'Or'(NE1,NE2)) :- !,
	elimNestedCasesInRHS(E1,NE1),
	elimNestedCasesInRHS(E2,NE2).
elimNestedCasesInRHS('Choice'(E),'Choice'(NE)) :- !,
	elimNestedCasesInRHS(E,NE).
elimNestedCasesInRHS(Exp,NewExp) :- elimCasesInExp(Exp,NewExp).

elimNestedCasesInBranches([],[]).
elimNestedCasesInBranches(['Branch'(Pat,Exp)|Cs],
	                  ['Branch'(Pat,NewExp)|NCs]) :-
	elimNestedCasesInRHS(Exp,NewExp),
	elimNestedCasesInBranches(Cs,NCs).


elimCasesInExp('Var'(V),'Var'(V)).
elimCasesInExp('Lit'(L),'Lit'(L)).
elimCasesInExp('Comb'(CT,CF,Args),'Comb'(CT,CF,NewArgs)) :-
	map2(elimCasesInExp,Args,NewArgs).
elimCasesInExp('Apply'(E1,E2),'Apply'(NE1,NE2)) :-
	elimCasesInExp(E1,NE1),
	elimCasesInExp(E2,NE2).
elimCasesInExp('Constr'(Vs,E),'Constr'(Vs,NE)) :-
	elimCasesInExp(E,NE).
elimCasesInExp('Or'(E1,E2),'Comb'('FuncCall',FNew,FVExps)) :-
	% elimCasesInExp(E1,NE1), no need to do that.
	% elimCasesInExp(E2,NE2),
	freeVarsInExp('Or'(E1,E2),FVs),
	map2(var2Exp,FVs,FVExps),
	genAuxFuncName(FNew),
	length(FVs,FNArity),
        elimNestedOrCases('Func'(FNew,FNArity,'Private',
				 'TVar'(0), % here I am too lazy
			                    % to compute the correct type
				 'Rule'(FVs,'Or'(E1,E2))),ElimF),
        addAuxFunction(ElimF).
elimCasesInExp('Case'(CT,CE,Branches),
	       'Comb'('FuncCall',FNew,[NCE|FVExps])) :-
	elimCasesInExp(CE,NCE),
	% elimCasesInBranches(Branches,NewBranches),
	freeVarsInBranches(Branches,FVs),
	map2(var2Exp,FVs,FVExps),
	genAuxFuncName(FNew),
	length(FVs,FVLength),
	FNArity is FVLength+1,
	newVarIndex(0,FVs,CVI),
	elimNestedOrCases('Func'(FNew,FNArity,'Private',
				 'TVar'(0), % here I am too lazy
			                    % to compute the correct type
				 'Rule'([CVI|FVs],'Case'(CT,'Var'(CVI),Branches))),
                         ElimF),
        addAuxFunction(ElimF).
elimCasesInExp('GuardedExpr'(Vs,E1,E2),'GuardedExpr'(Vs,NE1,NE2)) :-
	elimCasesInExp(E1,NE1),
	elimCasesInExp(E2,NE2).
elimCasesInExp('Choice'(E),'Choice'(NE)) :-
	elimCasesInExp(E,NE).
% preliminary solution (hack?): transform (Let [(x1,e1),...,(xn,en)] e) into
% (Constr2 [x1,...,xn] (letrec x1 e1 &>...&> letrec xn en &> e)
elimCasesInExp('Let'(Bindings,E),'Constr'(Vs,BindingList)) :-
	map2(bindingVar,Bindings,Vs),
	map2(elimCasesInBinding,Bindings,NBindings),
	elimCasesInExp(E,NE),
	letbindings2constr(NBindings,NE,BindingList).

bindingVar('(,)'(V,_),V).

letbindings2constr([],Exp,Exp).
letbindings2constr(['(,)'(X,E)|Bs],Exp,
		   'Comb'('FuncCall',Cond,['Comb'('FuncCall',LetRec,['Var'(X),E]),BsL])) :-
	atom_chars('letrec',LetRec),
        atom_chars('cond',Cond),
        letbindings2constr(Bs,Exp,BsL).
	
elimCasesInBinding('(,)'(V,E),'(,)'(V,NE)) :- elimCasesInExp(E,NE).

elimCasesInBranches([],[]).
elimCasesInBranches(['Branch'(Pat,Exp)|Cs],
	            ['Branch'(Pat,NewExp)|NCs]) :-
	elimCasesInExp(Exp,NewExp),
	elimCasesInBranches(Cs,NCs).

var2Exp(I,'Var'(I)).

% find a new variable index not occurring in a list of variables:
newVarIndex(Index,Vars,NewIndex) :-
	member(Index,Vars)
	 -> I1 is Index+1, newVarIndex(I1,Vars,NewIndex)
	  ; NewIndex=Index.

% generate new name for auxiliary function:
genAuxFuncName(FauxName) :-
	retract(newFunctionCounter(FNameS,FC)),
	FC1 is FC+1,
	asserta(newFunctionCounter(FNameS,FC1)),
	number_chars(FC,FCS),
	append(FNameS,"_caseor$auxfun_",FNameAux),
	append(FNameAux,FCS,FauxName).

% add new auxiliary function:
addAuxFunction(AF) :-
	retract(newAuxFunctions(AFs)),
	asserta(newAuxFunctions([AF|AFs])).


% compute the correct types of a list of functions w.r.t. to a list of already
% type correct functions:
computeCorrectType(_,[],[]).
computeCorrectType(AllFunData,['Func'(Name,Arity,Vis,_,'Rule'(Args,RHS))|Funs],
		   [NewFun|NewFuns]) :-
	%writeErr('Compute type of '),
	%atom_chars(NameA,Name), writeErr(NameA), nlErr,
	%writeFunTypes(AllFunData), nlErr,
	map2(var2vartype,Args,TEnv),
	typeExpr(RHS,AllFunData,TEnv,EType),
	tenvtype2funtype(TEnv,EType,Type),
	freevars2tvars(Type,0,_),
	NewFun='Func'(Name,Arity,Vis,Type,'Rule'(Args,RHS)),
	computeCorrectType([NewFun|AllFunData],Funs,NewFuns).
computeCorrectType(AllFunData,['Func'(Name,Arity,Vis,Type,'External'(EName))|Funs],
		   [NewFun|NewFuns]) :-
	NewFun='Func'(Name,Arity,Vis,Type,'External'(EName)),
	computeCorrectType([NewFun|AllFunData],Funs,NewFuns).

% for testing:
writeFunTypes([]).
writeFunTypes(['Func'(Name,_,_,_,_)|FDs]) :-
	atom_chars(NameA,Name), writeErr(NameA), writeErr(' '),
	writeFunTypes(FDs).
writeFunTypes(['Type'(Name,_,_,_)|FDs]) :-
	atom_chars(NameA,Name), writeErr(NameA), writeErr(' '),
	writeFunTypes(FDs).


tenvtype2funtype([],T,T).
tenvtype2funtype([type(_,T)|Ts],RT,'FuncType'(T,FT)) :-
	tenvtype2funtype(Ts,RT,FT).

var2vartype(V,type(V,_)).

getTypeFromTypeEnv([type(V,T)|_],V,T) :- !.
getTypeFromTypeEnv([_|TEnv],V,T) :- getTypeFromTypeEnv(TEnv,V,T).

typeExpr('Var'(V),_,TEnv,Type) :- getTypeFromTypeEnv(TEnv,V,Type), !.
typeExpr('Lit'('Intc'(_)),_,_,'TCons'("Int",[])) :- !.
typeExpr('Lit'('Floatc'(_)),_,_,'TCons'("Float",[])) :- !.
typeExpr('Lit'('Charc'(_)),_,_,'TCons'("Char",[])) :- !.
typeExpr('Comb'(_,CF,Args),Funs,TE,T) :-
	getTypeOfFunction(Funs,CF,FunType),
	typeExprs(Args,Funs,TE,FunType,T), !.
typeExpr('Apply'(E1,E2),Funs,TE,T2) :-
	typeExpr(E1,Funs,TE,FType),
	typeExpr(E2,Funs,TE,T1),
	(FType='FuncType'(T1,T2) -> true ;
	   writeErr('*** Illegal FlatCurry file: Type error (Apply):'),
	   nlErr,
	   writeErr('*** Term: '), writeErr(E1),
	   nlErr,
	   writeErr('*** Inferred type: '),
	   writeErr(FType), nlErr,
	   writeErr('*** Expected type: '),
	   writeErr('FuncType'(T1,T2)), nlErr,
	   !, fail), !.
typeExpr('Constr'(Vs,E),Funs,TE,T) :-
	map2(var2vartype,Vs,TVs),
	append(TVs,TE,VTE),
	typeExpr(E,Funs,VTE,T), !.
typeExpr('Or'(E1,E2),Funs,TE,T) :-
	typeExpr(E1,Funs,TE,T1),
	typeExpr(E2,Funs,TE,T2),
	(T1=T2 -> T=T1 ;
	   writeErr('*** Illegal FlatCurry file: Type error (Or):'),
	   nlErr,
	   writeErr('*** Term: '), writeErr('Or'(E1,E2)),
	   nlErr,
	   writeErr('*** Inferred type of first argument:  '),
	   writeErr(T1), nlErr,
	   writeErr('*** Inferred type of second argument: '),
	   writeErr(T2), nlErr,
	   !, fail), !.
typeExpr('Case'(_,CE,Branches),Funs,TE,T) :-
	typeExpr(CE,Funs,TE,CT),
	typeBranches(Branches,CT,Funs,TE,T), !.
typeExpr('GuardedExpr'(Vs,E1,E2),Funs,TE,T) :-
	map2(var2vartype,Vs,TVs),
	append(TVs,TE,VTE),
	typeExpr(E1,Funs,VTE,SType),
	typeExpr(E2,Funs,VTE,T),
	(SType='TCons'("Success",[]) -> true ;
	   writeErr('*** Illegal FlatCurry file: Type error (GuardedExpr):'),
	   nlErr,
	   writeErr('*** Term: '), writeErr(E1),
	   nlErr,
	   writeErr('*** Inferred type: '),
	   writeErr(SType), nlErr,
	   writeErr('*** Expected type: '),
	   writeErr('TCons'("Success",[])), nlErr,
	   !, fail), !.
typeExpr('Choice'(E),Funs,TE,T) :-
	typeExpr(E,Funs,TE,T), !.
% TODO: add Let case

typeExprs([],_,_,Type,Type).
typeExprs([E|Es],Funs,TE,'FuncType'(T1,T2),ResultType) :-
	typeExpr(E,Funs,TE,TA1),
	(TA1=T1 -> true ;
	   writeErr('*** Illegal FlatCurry file: Type error (FunArgs):'),
	   nlErr,
	   writeErr('*** Term: '), writeErr(E),
	   nlErr,
	   writeErr('*** Inferred type: '),
	   writeErr(TA1), nlErr,
	   writeErr('*** Expected type: '),
	   writeErr(T1), nlErr,
	   !, fail),
	typeExprs(Es,Funs,TE,T2,ResultType).

typeBranches([],_,_,_,_).
typeBranches(['Branch'('LPattern'(_),Exp)|Bs],CT,Funs,TE,T) :-
	typeExpr(Exp,Funs,TE,T),
	typeBranches(Bs,CT,Funs,TE,T).
typeBranches(['Branch'('Pattern'(Cons,Vs),Exp)|Bs],CT,Funs,TE,T) :-
	map2(var2vartype,Vs,TVs),
	append(TVs,TE,VTE),
	getTypeOfFunction(Funs,Cons,ConsType),
	map2(varIndex2VarExp,Vs,Vars),
	typeExprs(Vars,Funs,VTE,ConsType,CT), % type pattern arguments
	typeExpr(Exp,Funs,VTE,T),
	typeBranches(Bs,CT,Funs,TE,T).

getTypeOfFunction([],Name,_) :- % usually, this case should not occur!
	writeErr('WARNING: Type of function '),
	atom_chars(AName,Name), writeErr(AName),
	writeErr(' not found!'), nlErr.
getTypeOfFunction(['Func'(Name,_,_,FType,_)|_],Name,Type) :- !,
	tvars2freevars(FType,[],Type,_).
getTypeOfFunction(['Type'(TC,_,TVs,Conss)|_],Name,Type) :-
	getTypeOfCons(Conss,Name,ArgTypes), !,
	map2(addTVar,TVs,TTVs),
	constype2funtype(ArgTypes,'TCons'(TC,TTVs),FType),
	tvars2freevars(FType,[],Type,_).
getTypeOfFunction([_|Funs],Name,FunType) :-
	getTypeOfFunction(Funs,Name,FunType).

addTVar(V,'TVar'(V)).
varIndex2VarExp(V,'Var'(V)).

constype2funtype([],T,T).
constype2funtype([T|Ts],RT,'FuncType'(T,FT)) :-	constype2funtype(Ts,RT,FT).

getTypeOfCons(['Cons'(Name,_,_,ArgTypes)|_],Name,ArgTypes) :- !.
getTypeOfCons([_|Conss],Name,ArgTypes) :- getTypeOfCons(Conss,Name,ArgTypes).


% transform all 'TVar's into Prolog variables:
tvars2freevars('TVar'(I),TVs,V,NTVs) :-
	addTVar2FreeVar(I,TVs,V,NTVs).
tvars2freevars('FuncType'(T1,T2),TVs,'FuncType'(NT1,NT2),NTVs) :-
	tvars2freevars(T1,TVs,NT1,NTVs1),
	tvars2freevars(T2,NTVs1,NT2,NTVs).
tvars2freevars('TCons'(TC,Ts),TVs,'TCons'(TC,NTs),NTVs) :-
	tvars2freevarsOnList(Ts,TVs,NTs,NTVs).

tvars2freevarsOnList([],TVs,[],TVs).
tvars2freevarsOnList([T|Ts],TVs,[NT|NTs],NTVs) :-
	tvars2freevars(T,TVs,NT,TVs1),
	tvars2freevarsOnList(Ts,TVs1,NTs,NTVs).

addTVar2FreeVar(I,[],V,[(I,V)]). % new entry for first variable occurrence
addTVar2FreeVar(I,[(I,V)|TVs],V,[(I,V)|TVs]) :- !.
addTVar2FreeVar(I,[TV|TVs],V,[TV|NTVs]) :- addTVar2FreeVar(I,TVs,V,NTVs).

	       
% transform all Prolog variables into numbered 'TVar's:
freevars2tvars(V,I,I1) :-
	var(V), !,
	V='TVar'(I),
	I1 is I+1.
freevars2tvars('TVar'(_),I,I).
freevars2tvars('FuncType'(T1,T2),I,I2) :-
	freevars2tvars(T1,I,I1),
	freevars2tvars(T2,I1,I2).
freevars2tvars('TCons'(_,Ts),I,J) :-
	freevars2tvarsOnList(Ts,I,J).

freevars2tvarsOnList([],I,I).
freevars2tvarsOnList([T|Ts],I,J) :-
	freevars2tvars(T,I,I1),
	freevars2tvarsOnList(Ts,I1,J).


% compute the free variables in an expression (where variables are represented
% as integers):
freeVarsInExp('Var'(V),[V]).
freeVarsInExp('Lit'(_),[]).
freeVarsInExp('Comb'(_,_,Terms),V) :-
	freeVarsInExps(Terms,V).
freeVarsInExp('Apply'(F,X),V) :-
	freeVarsInExp(F,V1),
	freeVarsInExp(X,V2),
	union(V1,V2,V).
freeVarsInExp('Constr'(CVars,Exp),V) :-
	freeVarsInExp(Exp,VE),
	diff(VE,CVars,V).
freeVarsInExp('Or'(E1,E2),V) :-
	freeVarsInExp(E1,V1),
	freeVarsInExp(E2,V2),
	union(V1,V2,V).
freeVarsInExp('Case'(_,CE,Branches),V) :-
	freeVarsInExp(CE,V1),
	freeVarsInBranches(Branches,V2),
	union(V1,V2,V).
freeVarsInExp('GuardedExpr'(GVars,Con,Exp),V) :-
	freeVarsInExp(Con,V1),
	freeVarsInExp(Exp,V2),
	union(V1,V2,V12),
	diff(V12,GVars,V).
freeVarsInExp('Choice'(Exp),V) :-
	freeVarsInExp(Exp,V).
freeVarsInExp('Let'(Bindings,Exp),V) :-
	freeVarsInBindings(Bindings,BVs),
	freeVarsInExp(Exp,EVs),
	union(BVs,EVs,BEVs),
	map2(bindingVar,Bindings,Vs),
	diff(BEVs,Vs,V).

freeVarsInExps([],[]).
freeVarsInExps([E|Es],V) :-
	freeVarsInExp(E,V1),
	freeVarsInExps(Es,V2),
	union(V1,V2,V).

freeVarsInBindings([],[]).
freeVarsInBindings(['(,)'(_,E)|Bs],V) :-
	freeVarsInExp(E,V1),
	freeVarsInBindings(Bs,V2),
	union(V1,V2,V).

freeVarsInBranches([],[]).
freeVarsInBranches(['Branch'('Pattern'(_,PVs),E)|Bs],V) :-
	freeVarsInExp(E,V1),
	diff(V1,PVs,V2),
	freeVarsInBranches(Bs,V3),
	union(V2,V3,V).
freeVarsInBranches(['Branch'('LPattern'(_),E)|Bs],V) :-
	freeVarsInExp(E,V1),
	freeVarsInBranches(Bs,V2),
	union(V1,V2,V).


% does a (logical) variable occures in an expression (where variables
% are represented as logical variables)?
occursInExp(V,'Var'(X)) :- X==V.
occursInExp(V,'Comb'(_,_,Terms)) :- occursInExps(V,Terms).
occursInExp(V,'Apply'(F,X)) :- occursInExp(V,F) ; occursInExp(V,X).
occursInExp(V,'Constr'(_,Exp)) :- occursInExp(V,Exp).
occursInExp(V,'Or'(E1,E2)) :- occursInExp(V,E1) ; occursInExp(V,E2).
occursInExp(V,'Case'(_,CE,Branches)) :-
	occursInExp(V,CE) ; occursInBranches(V,Branches).
occursInExp(V,'GuardedExpr'(_,Con,Exp)) :-
	occursInExp(V,Con) ; occursInExp(V,Exp).
occursInExp(V,'Choice'(Exp)) :-	occursInExp(V,Exp).

occursInExps(V,[E|Es]) :- occursInExp(V,E) ; occursInExps(V,Es).

occursInBranches(V,['Branch'(_,E)|Bs]) :-
	occursInExp(V,E) ; occursInBranches(V,Bs).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Auxiliaries for handling dynamic data:

% Add new initialization directives for dynamic data (used for saving states):
addDynamicNameInfo(PredName,FileName) :-
	retract(dynamicPredNames(Ds)),
	asserta(dynamicPredNames([(PredName,FileName)|Ds])), !.

% Write initial info clause for dynamic predicates:
writeDynamicInfoClause((PredName,'')) :- !,
	writeClause((:- assertz(dynamicPredInfo(PredName,'')))).
writeDynamicInfoClause((PredName,FileName)) :-
	evalToken(Eval),
	DynInfoClause =
	  (dynamicPredInfo(PredName,RF) :-
		       nf(FileName,R,Eval,E1), boolEq(R,R,_,E1,_),
		       string2Atom(R,RA), checkDynamicAccessMethod(RA,RF),
		       retract((dynamicPredInfo(PredName,_) :- _)),
		       asserta(dynamicPredInfo(PredName,RF))),
	writeClause((:- assertz(DynInfoClause))).


% is the rhs of a rule a dynamic declaration?
isDynamicRuleDecl([],'Comb'('FuncCall',"Dynamic.dynamic",[]),'') :- !.
isDynamicRuleDecl([],'Comb'('FuncCall',"Dynamic.persistent",[S]),S) :- !.
isDynamicRuleDecl(Args,'Apply'(Exp,'Var'(A)),S) :-
	append(As,[A],Args),
	isDynamicRuleDecl(As,Exp,S).


% check the type of dynamic predicates (i.e., result type Dynamic and
% monomorphism restriction):
checkDynamicType(PredName,'FuncType'(T1,T2)) :- !,
	checkDynamicTypeForMonomorphism(PredName,T1),
	checkDynamicType(PredName,T2).
checkDynamicType(_,'TCons'("Dynamic.Dynamic",[])) :- !.
checkDynamicType(PredName,_) :-
	writeErr('ERROR: Dynamic predicate "'),
	writeErr(PredName),
	writeErr('" has not result type "Dynamic"!'),  nlErr,
	setFlcBug.

checkDynamicTypeForMonomorphism(PredName,'FuncType'(T1,T2)) :-
	checkDynamicTypeForMonomorphism(PredName,T1),
	checkDynamicTypeForMonomorphism(PredName,T2).
checkDynamicTypeForMonomorphism(PredName,'TCons'(_,Ts)) :-
	map1partial(checkDynamicTypeForMonomorphism(PredName),Ts).
checkDynamicTypeForMonomorphism(PredName,'TVar'(_)) :-
	writeErr('ERROR: Type of dynamic predicate "'),
	writeErr(PredName),
	writeErr('" contains type variable!'),  nlErr,
	setFlcBug.


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% translation of a defined function
writeFunc('Func'(_,_,_,_,'External'(_))). % no code for external functions

% special code for dynamic predicates:
writeFunc('Func'(Name,FArity,_Vis,Type,'Rule'(Args,Exp))) :-
	isDynamicRuleDecl(Args,Exp,DynDeclArg), !,
	flatName2Atom(Name,FName),
	checkDynamicType(FName,Type),
	atom_chars(FName,FNameS),
	append(_,[46|UQNameS],FNameS),
	append("$DYN_",UQNameS,DynNameS),
	atom_chars(DynName,DynNameS),
        (DynDeclArg = ''
	  -> DynType = 'Dynamic.Temporary'
	   ; exp2Term([],DynDeclArg,ENameTerm),
	     DynType = 'Dynamic.Persistent'),
        length(Xs,FArity),
	length(Ys,FArity),
	PYs =.. [DynName,DynType,0,0|Ys],
	append(Xs,['Dynamic.Dynamic'(PYs),E0,E],HeadArgs),
	Head =.. [FName|HeadArgs],
	genNfArgs(Xs,Ys,E0,E,Body),
	writeClause((Head :- Body)),
	DArity is FArity+2,
	writeClause((:- dynamic DynName/DArity)),
	(DynType = 'Dynamic.Temporary'
	 -> addDynamicNameInfo(DynName/DArity,'')
	  ; addDynamicNameInfo(DynName/DArity,ENameTerm)),
	nl, !.

writeFunc('Func'(Name,FArity,_Vis,_Type,'Rule'(FlatArgs,FlatExp))) :-
	flatName2Atom(Name,FName),
	retract(currentFunction(_)),
	asserta(currentFunction(FName)),
	% only for debugging:
	  %writeErr('*** Translating rule: '),
	  %writeErr(FName), writeErr(' '),
	  %writeErr(FlatArgs), writeErr(' = '),
	  %writeErr(FlatExp), nlErr,
	write(':- block '), writeq(FName), write('(?,'),
	writeNTimes(FArity,'?,'), write('-,?).'), nl,
	flatargs2var(FlatArgs,Env,Args),
	flatexp2var(Env,FlatExp,RHS),
	transExp(FName,'',Args,Args,RHS), nl.

genNfArgs([],[],E,E,true).
genNfArgs([X|Xs],[Y|Ys],E0,E,(nf(X,Y,E0,E1),Goal)) :-
	genNfArgs(Xs,Ys,E1,E,Goal).

% assign to each FlatCurry argument (an integer) in an argument list a new
% logic variable and unify second argument with this assignment, e.g.,
% flatargs2var([1,2,3],E,L) --> E=[(1,X),(2,Y),(3,Z)],L=[X,Y,Z]
flatargs2var([],[],[]).
flatargs2var([I|Is],[(I,Var)|E],[Var|Vars]) :- flatargs2var(Is,E,Vars).

% replace in a FlatCurry expression all variable indices by logic variables:
flatexp2var(Env,'Var'(I),'Var'(V)) :- !,
	getVarInEnv(I,Env,V).
flatexp2var(_,'Lit'(L),'Lit'(L)) :- !.
flatexp2var(Env,'Comb'(CT,CF,Exps),'Comb'(CT,CF,VarExps)) :- !,
	map2partial(flatexp2var(Env),Exps,VarExps).
flatexp2var(Env,'Apply'(E1,E2),'Apply'(VE1,VE2)) :- !,
	flatexp2var(Env,E1,VE1),
	flatexp2var(Env,E2,VE2).
flatexp2var(Env,'Constr'(VarIs,Exp),'Constr'(Vars,VarExp)) :- !,
	flatargs2var(VarIs,VarEnv,Vars),
	append(VarEnv,Env,Env1),
	flatexp2var(Env1,Exp,VarExp).
flatexp2var(Env,'Or'(E1,E2),'Or'(VE1,VE2)) :- !,
	flatexp2var(Env,E1,VE1),
	flatexp2var(Env,E2,VE2).
flatexp2var(Env,'Case'(CT,Exp,Cases),'Case'(CT,VarExp,VarCases)) :- !,
	flatexp2var(Env,Exp,VarExp),
	map2partial(flatcases2var(Env),Cases,VarCases).
flatexp2var(Env,'GuardedExpr'(VarIs,E1,E2),'GuardedExpr'(Vars,VE1,VE2)) :- !,
	flatargs2var(VarIs,VarEnv,Vars),
	append(VarEnv,Env,Env1),
	flatexp2var(Env1,E1,VE1),
	flatexp2var(Env1,E2,VE2).
flatexp2var(Env,'Choice'(Exp),'Choice'(VarExp)) :- !,
	flatexp2var(Env,Exp,VarExp).
flatexp2var(_,Expr,Expr) :-
	writeErr('ERROR in FlatCurry file: Unknown expression "'),
	writeErr(Expr),
	writeErr('" in function "'),
	currentFunction(FName), writeErr(FName),
	writeErr('"!'), nlErr,
	pleaseReport,
	write('\%ERROR in FlatCurry file: Unknown expression "'),
	write(Expr), write('" in function "'),
	write(FName), write('"!'), nl.

flatcases2var(Env,'Branch'('Pattern'(L,VarIs),Exp),
	          'Branch'('Pattern'(L,Vars),VarExp)) :-
	flatargs2var(VarIs,VarEnv,Vars),
	append(VarEnv,Env,Env1),
	flatexp2var(Env1,Exp,VarExp).
flatcases2var(Env,'Branch'('LPattern'(L),Exp),
	          'Branch'('LPattern'(L),VarExp)) :-
	flatexp2var(Env,Exp,VarExp).

getVarInEnv(_,[],_) :-
	writeErr('ERROR in FlatCurry file: '),
	writeErr('undeclared variable in function "'),
	currentFunction(FName), writeErr(FName),
	writeErr('"!'), nlErr,
	reportLiftBug,
	write('\%ERROR in FlatCurry file: '),
	write('undeclared variable in function "'),
	write(FName), write('"!'), nl.
getVarInEnv(I,[(I,Var)|_],Var) :- !.
getVarInEnv(I,[_|Env],Var) :- getVarInEnv(I,Env,Var).

	
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% write clauses with type (and evaluation) information

writeFTypeClause(ExtFuncs,Ops,'Func'(Name,_FArity,Vis,FlatType,_)) :-
	flatName2Atom(Name,FName),
	replaceTVarByLVar([],FlatType,_,FlatTypeP),
	getExternalNameFromVisibility(Name,Vis,EName),
	getFuncArity(FName,Arity),
	getPrologNameFromExtFuncs(FName,Arity,ExtFuncs,PrologName),
	getFixityFromOpList(FName,Ops,Fixity),
	writeClause(functiontype(FName,EName,Arity,PrologName,Fixity,FlatTypeP)).

getPrologNameFromExtFuncs(FName,Arity,ExtFuncs,PrologName) :-
	member((FName/Arity,PName),ExtFuncs), !,
	(append(_,[32|PrimNameS],PName)
	 -> atom_chars(PrologName,PrimNameS)
	  ; writeErr('ERROR: Specification of primitive function '),
	    deleteCostCenterInName(FName,FNameWOCC),
	    writeErr(FNameWOCC), writeErr('/'), writeErr(Arity),
	    writeErr(' not found!'), nlErr,
	    setFlcBug, fail).
getPrologNameFromExtFuncs(FName,_,_,FName).

getFixityFromOpList(FName,Ops,Fixity) :-
	atom_chars(FName,FNameS),
	member('Op'(FNameS,Type,Prec),Ops), !,
	transFixity2pl(Type,Prec,Fixity).
getFixityFromOpList(_,_,nofix).

transFixity2pl('InfixOp',P,infix(P)).
transFixity2pl('InfixlOp',P,infixl(P)).
transFixity2pl('InfixrOp',P,infixr(P)).

getExternalNameFromVisibility(Name,'Private',FName) :- atom_chars(FName,Name).
getExternalNameFromVisibility(Name,'Public',FName) :-
	append([_|_],[46|F],Name), !, atom_chars(FName,F).
getExternalNameFromVisibility(Name,'Public',FName) :- atom_chars(FName,Name).

getExternalNameFromTranslate(FName,Trans,EName) :-
	member(translate(EName,FName),Trans) -> true ; EName=FName.
%getExternalNameFromTranslate(FName,[],FName).
%getExternalNameFromTranslate(FName,[translate(EName,FName)|_],EName) :- !.
%getExternalNameFromTranslate(FName,[_|Translate],EName) :-
%	getExternalNameFromTranslate(FName,Translate,EName).

writeDTypeClause('Type'(TypeName,_Vis,TypeArgs,ConsExprs)) :-
	map2(index2tvar,TypeArgs,TypeArgExps),
	ResultType = 'TCons'(TypeName,TypeArgExps),
	writeDTypeClauses(ResultType,ConsExprs).

index2tvar(I,'TVar'(I)). % transform tvar index into type expression

writeDTypeClauses(_,[]).
writeDTypeClauses(ResultType,['Cons'(ConsName,Arity,Vis,ArgTypes)|Cs]) :-
	flatName2Atom(ConsName,Cons),
	append(ArgTypes,[ResultType],TypeL),
	typelist2flattype(TypeL,CType),
	replaceTVarByLVar([],CType,_,CTypeP),
	%getExternalNameFromTranslate(Cons,Translate,EName),
	getExternalNameFromVisibility(ConsName,Vis,EName),
	writeClause(constructortype(Cons,EName,Arity,CTypeP)),
	writeDTypeClauses(ResultType,Cs).

typelist2flattype([Type],Type) :- !.
typelist2flattype([T1|T2L],'FuncType'(T1,T2)) :-
	!,
	typelist2flattype(T2L,T2).


% remove all "TVar" constructors in a type expression by logic variables
% and replace FlatCurry names by atoms for better readability of
% the generated Prolog programs:
replaceTVarByLVar(Env,'TVar'(I),Env,TVar) :-
	getTVarInEnv(I,Env,TVar), !.
replaceTVarByLVar(Env,'TVar'(I),[(I,TVar)|Env],TVar).
replaceTVarByLVar(Env,'FuncType'(T1,T2),NewEnv,'FuncType'(TT1,TT2)) :-
	replaceTVarByLVar(Env,T1,Env1,TT1),
	replaceTVarByLVar(Env1,T2,NewEnv,TT2).
replaceTVarByLVar(Env,'TCons'(Name,Types),NewEnv,'TCons'(Cons,TTypes)) :-
	flatName2Atom(Name,Cons),
	replaceTVarByLVarL(Env,Types,NewEnv,TTypes).

replaceTVarByLVarL(Env,[],Env,[]).
replaceTVarByLVarL(Env,[Type|Types],NewEnv,[TType|TTypes]) :-
	replaceTVarByLVar(Env,Type,Env1,TType),
	replaceTVarByLVarL(Env1,Types,NewEnv,TTypes).

getTVarInEnv(I,[(I,TVar)|_],TVar) :- !.
getTVarInEnv(I,[_|Env],TVar) :- getTVarInEnv(I,Env,TVar).



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% translation of clauses for evaluation to head normal form

trans_hnf_current(CCs) :-
	findall(FE,functiontype2external(FE),ExtFuncs),
	trans_hnf([],ExtFuncs,CCs).

functiontype2external((FName/FArity,[65,32|ENameS])) :-
	functiontype(FName,_,FArity,EName,_,_),
	atom_chars(EName,ENameS).

trans_hnf(Funcs,ExtFuncs,CCs) :-
	writeClause((:- block hnf(?,?,-,?))),
	writeClause((hnf(T,H,E0,E) :- var(T), !, H=T, E0=E)),
	(compileWithSharing(variable)
	 -> map1(genVariableShareHnfClause,CCs)
	  ; (compileWithSharing(function)
	     -> map1(genFunctionShareHnfClause,CCs)
	      ; true)),
	map1(transhnf_clause,Funcs),
	map1(transExthnf,ExtFuncs),
	writeClause(hnf(T,T,E,E)), nl.

genVariableShareHnfClause(Suffix) :-
	appendAtom(share,Suffix,Share),
	Share_M =.. [Share,M],
	appendAtom(propagateShare,Suffix,PropShare),
	PropShare_HV_R =.. [PropShare,HV,R],
	writeClause((hnf(Share_M,R,E0,E) :- !, get_mutable(V,M),
                        (V='$eval'(Expr)
                         -> R=Expr, E0=E
                          ; hnf(V,HV,E0,E1),
			    ((nonvar(HV),functor(HV,success,0))
			      -> R=HV % no sharing for constraints
                               ; PropShare_HV_R,
				 update_mutable('$eval'(R),M)),
			    E1=E))).

genFunctionShareHnfClause(Suffix) :-
	appendAtom(share,Suffix,Share),
	Share_M =.. [Share,M],
	writeClause((hnf(Share_M,R,E0,E) :- !, get_mutable(V,M),
                        (V='$eval'(Expr)
                         -> R=Expr, E0=E
                          ; hnf(V,R,E0,E1),
			    update_mutable('$eval'(R),M),
			    E1=E))).

% generate hnf clause for an external function w.r.t. specified evaluation
% of arguments:
transExthnf((success/0,_)) :- !. % success is treated as a constructor
transExthnf((FName/0,"success")) :- !,
	% if somebody defines FName as (External success)
	writeClause((hnf(FName,success,E,E) :- !)).
transExthnf((FName/FArity,PName)) :-
	(append(_,[32|PrimNameS],PName)
	 -> atom_chars(PrimName,PrimNameS),
 	    transhnf_clause4func(FName,FArity,PrimName)
	  ; writeErr('ERROR: Specification of primitive function '),
	    deleteCostCenterInName(FName,FNameWOCC),
	    writeErr(FNameWOCC), writeErr('/'), writeErr(FArity),
	    writeErr(' not found!'), nlErr,
	    setFlcBug, fail).
transExthnf(_) :- !.

% hnf-clauses for external functions are generated by transExthnf
transhnf_clause('Func'(_,_,_,_,'External'(_))).
transhnf_clause('Func'(Name,Arity,_,_,'Rule'(FlatArgs,_))) :-
	flatName2Atom(Name,FName),
	length(FlatArgs,NumFA),
	(NumFA=Arity -> true
	              ; writeErr('INTERNAL ERROR: inconsistent arities'),
	                nlErr,
	                setFlcBug),
	transhnf_clause4func(FName,Arity,FName).
	
transhnf_clause4func(FName,FArity,PredName) :-
	length(Args,FArity),
	LHS =.. [FName|Args],
	append(Args,[R,E0,E],RArgs),
	PredCall =.. [PredName|RArgs],
	(compileWithDebug ->
	      Goal = (traceCall(LHS,Skip), PredCall, traceExit(LHS,R,E,Skip))
            ; Goal = PredCall),
	(compileWithFailPrint
	 -> writeClause((hnf(LHS,R,E0,E) :- !, (Goal; failprint(LHS,E0,E))))
	  ; writeClause((hnf(LHS,R,E0,E) :- !, Goal))).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% translation of clauses for higher-order application

transapply(Funcs) :-
	writeClause((:- block '$apply'(-,?,?,?,?), '$apply'(?,?,?,-,?))),
	map1(transapply_clause,Funcs),
	nl.

transapply_clause(FName/Arity) :-
	Arity>0, !,
	transapply_partial_clauses(FName,0,Arity),
	Arity1 is Arity-1,
	functor(F_Call,FName,Arity1),
	F_Call =.. [_|Args],
	append(Args,[X],ArgsX),
	F_Call_X =.. [FName|ArgsX],
	writeClause(('$apply'(F_Call,X,H,E0,E) :- hnf(F_Call_X,H,E0,E))).
transapply_clause(_).

transapply_partial_clauses(FName,K,Arity) :-
	K < Arity-1, !,
	functor(F_Call,FName,K),
	F_Call =.. [_|Args],
	append(Args,[X],ArgsX),
	F_Call_X =.. [FName|ArgsX],
	writeClause('$apply'(F_Call,X,F_Call_X,E,E)),
	K1 is K+1,
	transapply_partial_clauses(FName,K1,Arity).
transapply_partial_clauses(_,_,_).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% translation of clauses for solving equational constraints

trans_constrEq(Suffix) :-
	appendAtom(constrEq,Suffix,ConstrEq),
	appendAtom(constrEqHnf,Suffix,ConstrEqHnf),
	ConstrEqMode =.. [ConstrEq,?,?,?,-,?],
	writeClause((:- block ConstrEqMode)),
	ConstrEq_A_B_E0_E =.. [ConstrEq,A,B,success,E0,E],
	ConstrEqHnf_HA_HB_E2_E =.. [ConstrEqHnf,HA,HB,E2,E],
	(compileWithDebug ->
            writeClause((ConstrEq_A_B_E0_E :- traceCall(=:=(A,B),Skip),
	                                 hnf(A,HA,E0,E1), hnf(B,HB,E1,E2),
	                                 ConstrEqHnf_HA_HB_E2_E,
					 traceExit(=:=(A,B),success,E,Skip)))
          ; writeClause((ConstrEq_A_B_E0_E :- hnf(A,HA,E0,E1),hnf(B,HB,E1,E2),
	                                 ConstrEqHnf_HA_HB_E2_E))),
	nl,
	ConstrEqHnfMode =.. [ConstrEqHnf,?,?,-,?],
	writeClause((:- block ConstrEqHnfMode)),
	%writeClause((constrEq_hnf(X,H,E0,E) :- var(X), !, bind(X,H,E0,E))),
	%writeClause((constrEq_hnf(H,X,E0,E) :- var(X), !, bind(X,H,E0,E))),
	ConstrEqHnf_X_H_E0_E =.. [ConstrEqHnf,X,H,E0,E],
	ConstrEqHnf_H_X_E0_E =.. [ConstrEqHnf,H,X,E0,E],
	appendAtom(bindTryNf,Suffix,BindTryNf),
	BindTryNf_X_H_E0_E =.. [BindTryNf,X,H,E0,E],
	writeClause((ConstrEqHnf_X_H_E0_E :- var(X), !, BindTryNf_X_H_E0_E)),
	writeClause((ConstrEqHnf_H_X_E0_E :- var(X), !, BindTryNf_X_H_E0_E)),
	ConstrEqHnf_T1_T2_E0_E =.. [ConstrEqHnf,T1,T2,E0,E],
	writeClause((ConstrEqHnf_T1_T2_E0_E :- integer(T1), !, T1=T2, E0=E)),
	writeClause((ConstrEqHnf_T1_T2_E0_E :- float(T1), !, {T1=T2}, E0=E)), %CLP
	ConstrEqHnf_A_B_E0_E =.. [ConstrEqHnf,A,B,E0,E],
	appendAtom(genConstrEqHnfBody,Suffix,GenConstrEqHnfBody),
	GenConstrEqHnfBody_1_NA =.. [GenConstrEqHnfBody,1,NA,A,B,EqBody],
	writeClause((ConstrEqHnf_A_B_E0_E :-
		        functor(A,FA,NA), functor(B,FB,NB),
		        FA==FB, NA==NB, GenConstrEqHnfBody_1_NA,
		        hnf(EqBody,_,E0,E))),
	nl,
	GenConstrEqHnfBody_N_NA_Succ =.. [GenConstrEqHnfBody,N,NA,_,_,'success'],
	writeClause((GenConstrEqHnfBody_N_NA_Succ :- N>NA,!)),
	appendAtom('=:=',Suffix,Eq),
	Eq_ArgA_ArgB =.. [Eq,ArgA,ArgB],
	GenConstrEqHnfBody_N_NA_Eq =.. [GenConstrEqHnfBody,N,NA,A,B,Eq_ArgA_ArgB],
	writeClause((GenConstrEqHnfBody_N_NA_Eq :- N=NA, !,
		       arg(N,A,ArgA), arg(N,B,ArgB))),
	appendAtom('&',Suffix,ConcAnd),
	ConcAnd_Eq =.. [ConcAnd,Eq_ArgA_ArgB,RemBody],
	GenConstrEqHnfBody_N_NA_And =.. [GenConstrEqHnfBody,N,NA,A,B,ConcAnd_Eq],
	GenConstrEqHnfBody_N1_NA =.. [GenConstrEqHnfBody,N1,NA,A,B,RemBody],
	writeClause((GenConstrEqHnfBody_N_NA_And :-
		       arg(N,A,ArgA), arg(N,B,ArgB),
		       N1 is N+1, GenConstrEqHnfBody_N1_NA)),
	nl,
	% optimization: try to normalize the term before binding:
	BindTryNf_X_T_E0_E =.. [BindTryNf,X,T,E0,E],
	appendAtom(nf,Suffix,Nf),
	Nf_T_NT_E0_E1 =.. [Nf,T,NT,E0,E1],
	appendAtom(bindDirect,Suffix,BindDirect),
	appendAtom(bind,Suffix,Bind),
	BindDirect_X_NT_E1_E =.. [BindDirect,X,NT,E1,E],
	Bind_X_T_E0_E =.. [Bind,X,T,E0,E],
	writeClause((BindTryNf_X_T_E0_E :- Nf_T_NT_E0_E1,
	                       (nonvar(E1) -> BindDirect_X_NT_E1_E
                                            ; Bind_X_T_E0_E))),
	BindDirect_X_T_E0_E =.. [BindDirect,X,T,E0,E],
	appendAtom(occursNot,Suffix,OccursNot),
	OccursNot_X_T =.. [OccursNot,X,T],
	writeClause((BindDirect_X_T_E0_E :- var(T), !, X=T, E0=E)),
	writeClause((BindDirect_X_T_E0_E :- OccursNot_X_T, X=T, E0=E)),
	nl,
	Bind_X_T_E0_E =.. [Bind,X,T,E0,E],
	writeClause((Bind_X_T_E0_E :- var(T), !, X=T, E0=E)),
	writeClause((Bind_X_T_E0_E :- integer(T), !, X=T, E0=E)),
	writeClause((Bind_X_T_E0_E :- float(T), !, {X=T}, E0=E)), %CLP
	Bind_A_B_E0_E =.. [Bind,A,B,E0,E],
	appendAtom(occursNotArgs,Suffix,OccursNotArgs),
	OccursNotArgs_1_NB_A_B =.. [OccursNotArgs,1,NB,A,B],
	appendAtom(bindArgs,Suffix,BindArgs),
	BindArgs_1_NB_A_B_E0_E =.. [BindArgs,1,NB,A,B,E0,E],
	writeClause((Bind_A_B_E0_E :-
		        functor(B,FB,NB), OccursNotArgs_1_NB_A_B,
		        functor(A,FB,NB), BindArgs_1_NB_A_B_E0_E)),
	nl,
	OccursNotArgs_N_NA_A_B =.. [OccursNotArgs,N,NA,A,B],
	OccursNotArgs_N1_NA_A_B =.. [OccursNotArgs,N1,NA,A,B],
	OccursNot_A_ArgB =.. [OccursNot,A,ArgB],
	writeClause((OccursNotArgs_N_NA_A_B :- N>NA,!)),
	writeClause((OccursNotArgs_N_NA_A_B :- arg(N,B,ArgB),
		        OccursNot_A_ArgB,
		        N1 is N+1, OccursNotArgs_N1_NA_A_B)),
	nl,
	BindArgs_N_NA_A_B_E0_E =.. [BindArgs,N,NA,A,B,E0,E],
	Bind_ArgA_HArgB_E1_E2 =.. [Bind,ArgA,HArgB,E1,E2],
	BindArgs_N1_NA_A_B_E2_E =.. [BindArgs,N1,NA,A,B,E2,E],
	writeClause((BindArgs_N_NA_A_B_E0_E :- N>NA,!, E0=E)),
	writeClause((BindArgs_N_NA_A_B_E0_E :-
		        arg(N,A,ArgA), arg(N,B,ArgB),
		        hnf(ArgB,HArgB,E0,E1), Bind_ArgA_HArgB_E1_E2,
		        N1 is N+1, BindArgs_N1_NA_A_B_E2_E)),
	nl,
	OccursNot_X_Y =.. [OccursNot,X,Y],
	writeClause((OccursNot_X_Y :- var(Y), !, X\==Y)),
	OccursNotArgs_1_NY_X_Y =.. [OccursNotArgs,1,NY,X,Y],
	writeClause((OccursNot_X_Y :- functor(Y,FY,NY), constructortype(FY,_,NY,_),
		                      !, OccursNotArgs_1_NY_X_Y)),
	writeClause(OccursNot_X_Y),
	nl.

trans_constrEq_hnf(Cons/Arity) :-
	functor(CX,Cons,Arity),
	functor(CY,Cons,Arity),
	CX =.. [_|Xs],
	CY =.. [_|Ys],
	(Arity=0 -> writeClause((constrEq_hnf(CX,CY,E,E) :- !))
         ; gen_constrEq_hnf_body(Xs,Ys,Body),
	   writeClause((constrEq_hnf(CX,CY,E0,E) :- !, hnf(Body,_,E0,E)))).

gen_constrEq_hnf_body([X],[Y],(X=:=Y)).
gen_constrEq_hnf_body([X1,X2|Xs],[Y1,Y2|Ys],&((X1=:=Y1),Body)) :-
	gen_constrEq_hnf_body([X2|Xs],[Y2|Ys],Body).


transbind(Cons/Arity) :-
	functor(CX,Cons,Arity),
	functor(CY,Cons,Arity),
	CX =.. [_|Xs],
	CY =.. [_|Ys],
	gen_bind_body(Xs,Ys,E0,E,HnfBody),
	gen_bind_occ_body(X,Ys,(X=CX,HnfBody),occursNot,Body),
	writeClause((bind(X,CY,E0,E) :- !, Body)).

gen_bind_body([],[],E,E,true).
gen_bind_body([X|Xs],[Y|Ys],E0,E,(hnf(Y,HY,E0,E1),bind(X,HY,E1,E2),Body)) :-
	gen_bind_body(Xs,Ys,E2,E,Body).

gen_bind_occ_body(_,[],RestBody,_,RestBody).
gen_bind_occ_body(X,[Y|Ys],RestBody,OccursNot,(OccursNot_X_Y,Body)) :-
	OccursNot_X_Y =.. [OccursNot,X,Y],
	gen_bind_occ_body(X,Ys,RestBody,OccursNot,Body).

transocc_not(OccursNot,Cons/Arity) :-
	functor(CY,Cons,Arity),
	CY =.. [_|Ys],
	gen_bind_occ_body(X,Ys,true,OccursNot,Body),
	OccursNot_X_CY =.. [OccursNot,X,CY],
	writeClause((OccursNot_X_CY :- !, Body)).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% translation of clauses for (Boolean) strict equality test
	
trans_boolEq(Suffix) :-
	appendAtom(boolEq,Suffix,BoolEq),
	appendAtom(boolEqHnf,Suffix,BoolEqHnf),
	BoolEqMode =.. [BoolEq,?,?,?,-,?],
	writeClause((:- block BoolEqMode)),
	BoolEq_A_B_R_E0_E =.. [BoolEq,A,B,R,E0,E],
	BoolEqHnf_HA_HB_R_E2_E =.. [BoolEqHnf,HA,HB,R,E2,E],
	writeClause((BoolEq_A_B_R_E0_E :- hnf(A,HA,E0,E1),hnf(B,HB,E1,E2),
	                                  BoolEqHnf_HA_HB_R_E2_E)),
	nl,
	BoolEqHnfMode1 =.. [BoolEqHnf,-,?,?,?,?],
	BoolEqHnfMode2 =.. [BoolEqHnf,?,-,?,?,?],
	BoolEqHnfMode3 =.. [BoolEqHnf,?,?,?,-,?],
	writeClause((:- block BoolEqHnfMode1, BoolEqHnfMode2, BoolEqHnfMode3)),
	BoolEqHnf_A_B_R_E0_E =.. [BoolEqHnf,A,B,R,E0,E],
	writeClause((BoolEqHnf_A_B_R_E0_E :-
		       number(A), !, (A==B->R='True';R='False'), E0=E)),
	appendAtom(genBoolEqHnfBody,Suffix,GenBoolEqHnfBody),
	GenBoolEqHnfBody_1_NA =.. [GenBoolEqHnfBody,1,NA,A,B,SeqBody],
	writeClause((BoolEqHnf_A_B_R_E0_E :-
		       functor(A,FA,NA),functor(B,FB,NB),
		       ((FA==FB,NA==NB,GenBoolEqHnfBody_1_NA)
		          -> hnf(SeqBody,R,E0,E)
		           ; R='False',E0=E))),
	nl,
	GenBoolEqHnfBody_N_NA_True =.. [GenBoolEqHnfBody,N,NA,_,_,'True'],
	writeClause((GenBoolEqHnfBody_N_NA_True :- N>NA,!)),
	appendAtom('==',Suffix,Eq),
	Eq_ArgA_ArgB =.. [Eq,ArgA,ArgB],
	appendAtom('&&',Suffix,BoolAnd),
	BoolAnd_Eq =.. [BoolAnd,Eq_ArgA_ArgB,RemBody],
	GenBoolEqHnfBody_N_NA_And =.. [GenBoolEqHnfBody,N,NA,A,B,BoolAnd_Eq],
	GenBoolEqHnfBody_N1_NA =.. [GenBoolEqHnfBody,N1,NA,A,B,RemBody],
	writeClause((GenBoolEqHnfBody_N_NA_And :-
		       arg(N,A,ArgA), arg(N,B,ArgB),
		       N1 is N+1, GenBoolEqHnfBody_N1_NA)),
	nl.


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% translation of clauses for evaluation to normal form
	
trans_nf(Suffix) :-
	appendAtom(nf,Suffix,Nf),
	appendAtom(nfHnf,Suffix,NfHnf),
	NfMode =.. [Nf,?,?,-,?],
	writeClause((:- block NfMode)),
	Nf_A_V_E0_E =.. [Nf,A,V,E0,E],
	NfHnf_HA_V_E1_E =.. [NfHnf,HA,V,E1,E],
	writeClause((Nf_A_V_E0_E :- hnf(A,HA,E0,E1), NfHnf_HA_V_E1_E)),
	nl,
	NfHnfMode =.. [NfHnf,?,?,-,?],
	writeClause((:- block NfHnfMode)),
	NfHnf_T_V_E0_E =.. [NfHnf,T,V,E0,E],
	writeClause((NfHnf_T_V_E0_E :- var(T), !, V=T, E0=E)),
	NfHnf_A_A_E0_E =.. [NfHnf,A,A,E0,E],
	writeClause((NfHnf_A_A_E0_E :- number(A), !, E0=E)),
	NfHnf_A_R_E0_E =.. [NfHnf,A,R,E0,E],
	appendAtom(nfHnfArgs,Suffix,NfHnfArgs),
	NfHnfArgs_1_NA_A_R_E0_E =.. [NfHnfArgs,1,NA,A,R,E0,E],
	writeClause((NfHnf_A_R_E0_E :-
		       functor(A,FA,NA), functor(R,FA,NA),
		       NfHnfArgs_1_NA_A_R_E0_E)),
	nl,
	NfHnfArgs_N_NA_A_R_E0_E =.. [NfHnfArgs,N,NA,A,R,E0,E],
	writeClause((NfHnfArgs_N_NA_A_R_E0_E :- N>NA,!,E0=E)),
	Nf_ArgA_ArgR_E0_E1 =.. [Nf,ArgA,ArgR,E0,E1],
	NfHnfArgs_N1_NA_A_R_E1_E =.. [NfHnfArgs,N1,NA,A,R,E1,E],
	writeClause((NfHnfArgs_N_NA_A_R_E0_E :-
		       arg(N,A,ArgA), arg(N,R,ArgR),
		       Nf_ArgA_ArgR_E0_E1,
		       N1 is N+1, NfHnfArgs_N1_NA_A_R_E1_E)),
	nl.


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% translation of clauses for create, propagate, and dereference sharing
	
genMakeFunctionShare(Suffix) :-
	appendAtom(makeShare,Suffix,MakeShare),
	appendAtom(share,Suffix,Share),
	Share_M =.. [Share,M],
	MakeShare_X_ShM =.. [MakeShare,X,Share_M],
	writeClause((MakeShare_X_ShM :- create_mutable(X,M))),
	nl.

transpropshar(Suffix) :-
	appendAtom(propagateShare,Suffix,PropShare),
	appendAtom(propagateShareArg,Suffix,PropShareArg),
	PropShare_X_X =.. [PropShare,X,X],
	writeClause((PropShare_X_X :- var(X), !)),
	PC1=partcall(N,F,PArg),
	PC2=partcall(N,F,SPArg),
	PropShare_PC_PC =.. [PropShare,PC1,PC2],
	writeClause((PropShare_PC_PC :- !, map2(PropShareArg,PArg,SPArg))),
	IP='Ports.internalPort'(_,_,_,_),
	PropShare_IP_IP =.. [PropShare,IP,IP],
	writeClause((PropShare_IP_IP :- !)),
	Str='$stream'(_),
	PropShare_Str_Str =.. [PropShare,Str,Str],
	writeClause((PropShare_Str_Str :- !)),
	writeClause((PropShare_X_X :- number(X),!)),
	PropShare_A_R =.. [PropShare,A,R],
	appendAtom(propagateShareArgs,Suffix,PropShareArgs),
	PropShareArgs_1_NA_A_R =.. [PropShareArgs,1,NA,A,R],
	writeClause((PropShare_A_R :-
		       functor(A,FA,NA), functor(R,FA,NA),
		       PropShareArgs_1_NA_A_R)),
	nl,
	PropShareArg_ArgA_ArgR =.. [PropShareArg,ArgA,ArgR],
	PropShareArgs_N_NA_A_R =.. [PropShareArgs,N,NA,A,R],
	PropShareArgs_N1_NA_A_R =.. [PropShareArgs,N1,NA,A,R],
	writeClause((PropShareArgs_N_NA_A_R :- N>NA,!)),
	writeClause((PropShareArgs_N_NA_A_R :-
		       arg(N,A,ArgA), arg(N,R,ArgR),
		       PropShareArg_ArgA_ArgR,
		       N1 is N+1, PropShareArgs_N1_NA_A_R)),
	nl,
	PropShareArg_A_A =.. [PropShareArg,A,A],
	writeClause((PropShareArg_A_A :- nonvar(A),
		              functor(A,'Ports.internalPort',4), !)),
	PropShareArg_A_SA =.. [PropShareArg,A,SA],
	appendAtom(makeShare,Suffix,MakeShare),
	MakeShare_A_SA =.. [MakeShare,A,SA],
	writeClause((PropShareArg_A_SA :- MakeShare_A_SA)),
	nl,
	MakeShare_X_X =.. [MakeShare,X,X],
	appendAtom(share,Suffix,Share),
	Share_M =.. [Share,M],
	MakeShare_X_Y =.. [MakeShare,X,Y],
	MakeShare_X_ShM =.. [MakeShare,X,Share_M],
	writeClause((MakeShare_X_X :- var(X), !)),
	writeClause((MakeShare_X_Y :- atomic(X), !, (functiontype(X,_,0,_,_,_) -> create_mutable(X,M), Y=Share_M ; Y=X))),
	writeClause((MakeShare_X_X :- X=Share_M, !)),
	writeClause((MakeShare_X_ShM :- create_mutable(X,M))),
	nl.

trans_deref(CCs) :-
	write('% dereference a term, i.e., remove all top-level sharing structures:'),
	nl,
	writeClause((deref(R,V) :- var(R), !, V=R)),
	map1(trans_derefClause,CCs),
	writeClause(deref(R,R)),
	nl,
	write('% completely dereference a term, i.e., remove all sharing structures'),
	nl,
	write('% also inside subterms:'), nl,
	writeClause((derefAll(R,V) :- var(R), !, V=R)),
	map1(trans_derefAllClause,CCs),
	writeClause((derefAll(R,V) :- R=..[C|As], map2(derefAll,As,Vs), V=..[C|Vs])),
	nl.

trans_derefClause(Suffix) :-
	appendAtom(share,Suffix,Share),
	Share_M =.. [Share,M],
	writeClause((deref(Share_M,V) :- !,
		        get_mutable(E,M), (E='$eval'(R) -> V=R ; deref(E,V)))).

trans_derefAllClause(Suffix) :-
	appendAtom(share,Suffix,Share),
	Share_M =.. [Share,M],
	writeClause((derefAll(Share_M,V) :- !,
	       get_mutable(E,M), (E='$eval'(R) -> derefAll(R,V) ; derefAll(E,V)))).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% translation of expressions:
% here we assume that all cases and ors occur only outermost and not
% at nested inner positions
transExp(FName,Aux,Patterns,Vars,'Case'(CType,CaseExp,Cases)) :- !,
	(CaseExp='Var'(X)
	 -> % case expression is variable:
	    selectVar(X,Vars,I,VarsX),
	    % check whether the case variable occurs in any case branch
	    (occursInBranches(X,Cases) -> NewVars=Vars ; NewVars=VarsX),
	    CaseTerm=X, UMShares=[], ShareCases=Cases,
	    number_chars(I,IS), atom_chars(IA,IS) % transform number into atom
	  ; % case expression is complex (not a variable)
	    (compileWithSharing(variable)
	     -> %exp2ShareExpr(CaseExp,UMShares,SCExp),
		getSharedVarsOfCaseArg('Case'(CType,CaseExp,Cases),UMShares),
		insertShare(CaseExp,UMShares,SCExp),
		map2partial(replaceShareVars(UMShares),Vars,NewVars),
		exp2Term(NewVars,SCExp,CaseTerm),
		map2partial(insertShareInBranch(UMShares),Cases,ShareCases)
	      ; (compileWithSharing(function)
	         -> exp2FuncShareTerm(Vars,CaseExp,UMShares,CaseTerm),
		    NewVars=Vars, ShareCases=Cases
	          ; exp2Term(Vars,CaseExp,CaseTerm),
		    UMShares=[], NewVars=Vars, ShareCases=Cases)),
	    IA='ComplexCase'  % naming needs to be improved!
	),
	atom_chars(FName,FNameS), getCostCenterOfName(FNameS,CC),
	map2partial(addSuffix2MakeShare(CC),UMShares,UShares),
	append(Patterns,[H,E0,E],LArgs),
	LHS =.. [FName|LArgs],
	appendAtom(FName,Aux,FNaux),
	appendAtom(FNaux,'_',FNameUS),
	appendAtom(FNameUS,IA,F_I),
	append(NewVars,[H,E1,E],F_I_Args),
	F_I_Call =.. [F_I,Y|F_I_Args],
	writeClauseWithInitGoals(LHS,UShares,(hnf(CaseTerm,Y,E0,E1), F_I_Call)), nl,
	length(NewVars,LenNewVars), NumVars is LenNewVars+1,
	write(':- block '), writeq(F_I),
	(CType='Rigid'
	 -> write('(-,'),
	    writeNTimes(NumVars,'?,'),
	    write('?,?),'),
	    writeq(F_I), write('(?,'),
	    writeNTimes(NumVars,'?,'), write('-,?).'), nl
	  ; % CType='Flex'
	    write('(?,'),
	    writeNTimes(NumVars,'?,'),
	    write('-,?).'), nl),
	transCases(F_I,NewVars,ShareCases).

transExp(FName,Aux,Patterns,Vars,'Or'(Exp1,Exp2)) :- !,
	(((Exp1 = 'Case'(_,_,_) ; Exp2 = 'Case'(_,_,_)),
	  pakcsrc(warnoverlapping,yes))
	 -> writeErrNQ('WARNING: function "'),
	    currentFunction(FuncName), writeErrNQ(FuncName),
	    writeErrNQ('" is non-deterministic due to non-trivial overlapping rules!'),
	    nlErrNQ
	  ; true),
	transExp(FName,Aux,Patterns,Vars,Exp1),
	transExp(FName,Aux,Patterns,Vars,Exp2).

transExp(FName,Aux,Patterns,Vars,'Choice'(Exp)) :- !,
	writeErr('ERROR: "'),
	writeErr(FName),
	writeErr(' eval choice" not yet supported!'),
	nlErr,
	transExp(FName,Aux,Patterns,Vars,Exp),
	setFlcBug.

transExp(FName,_Aux,Patterns,Vars,Exp) :-
	append(Patterns,[H,E0,E],LArgs),
	LHS =.. [FName|LArgs],
	(compileWithSharing(variable)
	 -> exp2ShareExpr(Exp,UMShares,SExp),
            map2partial(replaceShareVars(UMShares),Vars,NewVars),
	    exp2Term(NewVars,SExp,Term)
	  ; (compileWithSharing(function)
	     -> exp2FuncShareTerm(Vars,Exp,UMShares,Term)
	      ; exp2Term(Vars,Exp,Term), UMShares=[])),
	atom_chars(FName,FNameS), getCostCenterOfName(FNameS,CC),
	map2partial(addSuffix2MakeShare(CC),UMShares,UShares),
	(isConstructorRooted(Exp)
	 -> % right-hand side is constructor-rooted -> no hnf call necessary:
            H=Term,
	    (UShares=[]
		-> E0=E, writeClause(LHS)
		 ; writeClauseWithInitGoals(LHS,UShares,E0=E))
	 ; ((var(Term) ; compileWithDebug ; compileWithFailPrint
	               ; (separateCompilation, \+ localFunCall(FName,Exp)))
	    -> writeClauseWithInitGoals(LHS,UShares,hnf(Term,H,E0,E))
             ; % if the right-hand side is operation-rooted ->
	       % generate direct call to root predicate:
  	       functor(Term,TF,TA),
	       Term =.. [TF|TArgs],
	       append(TArgs,[H,E0,E],TArgsH),
	       externalFuncs(EFs),
	       (member((TF/TA,PrimName),EFs)
	        -> append(_,[32|TPS],PrimName), atom_chars(TP,TPS) ; TP=TF),
	       TPCall =.. [TP|TArgsH],
               writeClauseWithInitGoals(LHS,UShares,TPCall) )).

localFunCall(_,'Comb'('FuncCall',"success",_)) :-
	!. % since we implement "success" as a constructor, thus we don't need hnf for it
localFunCall(DefFuncName,'Comb'('FuncCall',FName,_)) :-
	atom_chars(DefFuncName,DefFuncNameS),
	fromSameModule(DefFuncNameS,FName), !.
localFunCall(DefFuncName,'Constr'(_,Exp)) :- % since we ignore Constr in the code:
	localFunCall(DefFuncName,Exp).

fromSameModule(F1,F2) :-
	append(Mod,[46|_],F1),
	append(Mod,[46|_],F2), !.
fromSameModule(F1,F2) :-
	\+ append(_,[46|_],F1),
	\+ append(_,[46|_],F2), % both are prelude operations
	!.

isConstructorRooted('Lit'(_)).
isConstructorRooted('Comb'('ConsCall',_,_)).
% we implement "success" as a constructor of type "Success":
isConstructorRooted('Comb'('FuncCall',"success",[])).
isConstructorRooted('Constr'(_,Exp)) :- % since we ignore Constr in the code:
	isConstructorRooted(Exp).

	
transCases(_,_,[]).
transCases(F_I,VarsX,['Branch'('Pattern'(Name,ConsVars),Exp)|Cases]) :-
	flatName2Atom(Name,Cons),
	Pattern =.. [Cons|ConsVars],
	append(ConsVars,VarsX,NewVars),
	(number(Cons) -> number_chars(Cons,ConsS),
	                 atom_chars(ConsA,ConsS)  % transform number into atom
		       ; ConsA=Cons),
	appendAtom('_',ConsA,Aux),
	transExp(F_I,Aux,[Pattern|VarsX],NewVars,Exp),
	transCases(F_I,VarsX,Cases).
transCases(F_I,VarsX,['Branch'('LPattern'(Lit),Exp)|Cases]) :-
	transCaseLit2Cons(Lit,Cons),
	(number(Cons) -> number_chars(Cons,ConsS),
	                 atom_chars(ConsA,ConsS)  % transform number into atom
		       ; ConsA=Cons),
	appendAtom('_',ConsA,Aux),
	transExp(F_I,Aux,[Cons|VarsX],VarsX,Exp),
	transCases(F_I,VarsX,Cases).

% extract constructor of literal:
transCaseLit2Cons('Intc'(I),I) :- !.
transCaseLit2Cons('Floatc'(F),F) :- !.
transCaseLit2Cons('Charc'(N),C) :- !, char_int(C,N).
transCaseLit2Cons(Lit,Lit) :-
	writeErr('ERROR in FlatCurry file: Illegal argument "'),
	writeErr(Lit),
	writeErr('" in case branch!'), nlErr,
	pleaseReport.

% translate FlatCurry expression into corresponding Prolog term:
% (first argument are the currently visible variables, only used
% to check inconsistencies in FlatCurry)
exp2Term(Vars,'Var'(V),V) :- !,
	(memberEq(V,Vars) -> true
	  ; writeErr('ERROR in FlatCurry file in function "'),
	    currentFunction(FuncName), writeErr(FuncName),
	    writeErr('":'), nlErr,
	    writeErr('variable "'),
	    writeErr(V),
	    writeErr('" does not occur in left-hand side.'),
	    nlErr,
	    reportLiftBug,
	    write('\%ERROR in FlatCurry file: variable "'),
	    write(V),
	    write('" does not occur in left-hand side!'), nl).
exp2Term(_,'Lit'('Intc'(I)),I) :- !.
exp2Term(_,'Lit'('Floatc'(F)),F) :- !.
exp2Term(_,'Lit'('Charc'(N)),C) :- !, char_int(C,N).
exp2Term(_,'Lit'('Ident'(S)),A) :- !,
	writeErr('ERROR in FlatCurry file in function "'),
	currentFunction(FuncName), writeErr(FuncName),
	writeErr('":'), nlErr,
	flatName2Atom(S,A),
	writeErr('Expression "'),
	writeErr('Lit'('Ident'(A))),
	writeErr('" should not occur in FlatCurry expressions.'),
	nlErr,
	pleaseReport,
	write('\%ERROR in FlatCurry file: Expression "'),
	write('Lit'('Ident'(A))),
	write('" should not occur in FlatCurry expressions.'),
	nl.
exp2Term(Vars,'Comb'(CombType,NameS,Exprs),Term) :- !,
	flatName2Atom(NameS,Name),
	checkForDeprecatedFunction(Name),
	checkForDynamicConstructor(Name),
	checkForTupleArity(NameS),
	map2partial(exp2Term(Vars),Exprs,Terms),
	length(Terms,TArity),
	(getConsArity(Name,Arity)
	 -> Missing=0 % constructors are not represented by partial applications
	 ; (getFuncArity(Name,Arity) -> Missing is Arity-TArity ; Missing=0)),
	(Missing=0 -> Term =.. [Name|Terms] ;
         (Missing>0 -> rev(Terms,RevTerms), Term = partcall(Missing,Name,RevTerms)
	             ; writeErr('INTERNAL COMPILER ERROR: over-application occured!'),
	               nlErr,
	               writeErr('Function: '),
	               currentFunction(QFunc), writeErr(QFunc), nlErr,
	               writeErr('Expression: '),
	               writeErr('Comb'(CombType,NameS,Exprs)), nlErr,
	               setFlcBug)).

% The following translation does not work for top-level constructor
% partial applications:
%exp2Term(Vars,'Apply'(F,X),Term) :- F='Comb'('PartCall',_,_), !,
%	exp2Term(Vars,F,FT), exp2Term(Vars,X,XT),
%	FT =.. [Func|Terms], append(Terms,[XT],TermsXT),
%	Term =.. [Func|TermsXT].
exp2Term(Vars,'Apply'(F,X),apply(FT,XT)) :- !,
	exp2Term(Vars,F,FT), exp2Term(Vars,X,XT).
exp2Term(Vars,'GuardedExpr'(LocalVars,Con,Exp),cond(CT,ET)) :- !,
	append(Vars,LocalVars,NewVars),
	exp2Term(NewVars,Con,CT), exp2Term(NewVars,Exp,ET).
exp2Term(Vars,'Constr'(LocalVars,Exp),ET) :- !,
	append(Vars,LocalVars,NewVars),
	exp2Term(NewVars,Exp,ET).
exp2Term(_,Expr,'***unknown expression***') :-
	writeErr('ERROR in FlatCurry file: Unknown expression "'),
	writeErr(Expr),
	writeErr('" in FlatCurry file!'), nlErr,
	pleaseReport.


% implement sharing scheme for variable sharing according to [Antoy/Hanus FroCos'2000]
% replace multiple variables by new variables connected to the old
% ones via a makeShare goal:
exp2ShareExpr(Exp,Shares,SExp) :-
	getSharedVars(Exp,Shares),
	insertShare(Exp,Shares,SExp).

insertShare('Var'(V),Shares,Share) :- !,
	varToShare(V,Shares,Share).
insertShare('Lit'(L),_,'Lit'(L)) :- !.
insertShare('Comb'(CT,Name,Exprs),Shares,'Comb'(CT,Name,SExprs)) :- !,
	insertShares(Exprs,Shares,SExprs).
insertShare('Apply'(F,X),Shares,'Apply'(SF,SX)) :- !,
	insertShare(F,Shares,SF), insertShare(X,Shares,SX).
insertShare('GuardedExpr'(Vars,Con,Exp),Shares,
		  'GuardedExpr'(NewVars,SCon,SExp)) :- !,
        map2partial(replaceShareVars(Shares),Vars,NewVars),
	insertShare(Con,Shares,SCon),
	insertShare(Exp,Shares,SExp).
insertShare('Constr'(Vars,Exp),Shares,'Constr'(NewVars,SExp)) :- !,
        map2partial(replaceShareVars(Shares),Vars,NewVars),
	insertShare(Exp,Shares,SExp).
insertShare('Choice'(Exp),Shares,'Choice'(SExp)) :- !,
	insertShare(Exp,Shares,SExp).
insertShare('Case'(CType,Exp,Branches),Shares,'Case'(CType,SExp,SBranches)) :-
	!,
	insertShare(Exp,Shares,SExp),
	map2partial(insertShareInBranch(Shares),Branches,SBranches).
insertShare('Or'(Exp1,Exp2),Shares,'Or'(SExp1,SExp2)) :-
	!,
	insertShare(Exp1,Shares,SExp1),
	insertShare(Exp2,Shares,SExp2).
insertShare(Exp,_,Exp) :-
	writeErr('INTERNAL ERROR in "insertShare"!'),
	nlErr,
	writeErr('Unknown expression in FlatCurry file:'),
	nlErr,
	writeErr(Exp),
	nlErr,
	pleaseReport.

insertShareInBranch(Shares,'Branch'(Pattern,Exp),'Branch'(Pattern,SExp)) :-
        insertShare(Exp,Shares,SExp).

insertShares([],_,[]).
insertShares([T|Ts],Shares,[ST|STs]) :-
	insertShare(T,Shares,ST),
	insertShares(Ts,Shares,STs).

varToShare(V,[],'Var'(V)).
varToShare(V,[makeShare(W,NV)|_],'Var'(NV)) :- V==W, !.
varToShare(V,[_|Shares],Share) :- varToShare(V,Shares,Share).

replaceShareVars([],V,V).
replaceShareVars([makeShare(V,NV)|_],W,NV) :- V==W, !.
replaceShareVars([_|UVs],W,NV) :- replaceShareVars(UVs,W,NV).

% implement sharing scheme for function sharing (classical technique):
exp2TermInMakeShare(Vars,makeShare(Exp,V),makeShare(Term,V)) :-
	exp2Term(Vars,Exp,Term).

exp2FuncShareTerm(Vars,Exp,Shares,Term) :-
	exp2FuncShareTerm(top,Vars,Exp,SharesR,Term),
	rev(SharesR,Shares).

exp2FuncShareTerm(_,Vars,'Var'(V),[],V) :- !,
	(memberEq(V,Vars) -> true
	  ; writeErr('ERROR in FlatCurry file in function "'),
	    currentFunction(FuncName), writeErr(FuncName),
	    writeErr('":'), nlErr,
	    writeErr('variable "'),
	    writeErr(V),
	    writeErr('" does not occur in left-hand side.'),
	    nlErr,
	    reportLiftBug,
	    write('\%ERROR in FlatCurry file: variable "'),
	    write(V),
	    write('" does not occur in left-hand side!'), nl).
exp2FuncShareTerm(_,_,'Lit'('Intc'(I)),[],I) :- !.
exp2FuncShareTerm(_,_,'Lit'('Floatc'(F)),[],F) :- !.
exp2FuncShareTerm(_,_,'Lit'('Charc'(N)),[],C) :- !, char_int(C,N).
exp2FuncShareTerm(_,_,'Lit'('Ident'(S)),[],A) :- !,
	writeErr('ERROR in FlatCurry file in function "'),
	currentFunction(FuncName), writeErr(FuncName),
	writeErr('":'), nlErr,
	flatName2Atom(S,A),
	writeErr('Expression "'),
	writeErr('Lit'('Ident'(A))),
	writeErr('" should not occur in FlatCurry expressions.'),
	nlErr,
	pleaseReport,
	write('\%ERROR in FlatCurry file: Expression "'),
	write('Lit'('Ident'(A))),
	write('" should not occur in FlatCurry expressions.'),
	nl.
exp2FuncShareTerm(Level,Vars,'Comb'(CombType,NameS,Exprs),NewShares,NewTerm) :- !,
	exp2FuncShareTerms(Vars,Exprs,Shares,Terms),
	flatName2Atom(NameS,Name),
	checkForDeprecatedFunction(Name),
	checkForDynamicConstructor(Name),
	checkForTupleArity(NameS),
	length(Terms,TArity),
	(getConsArity(Name,Arity)
	 -> Missing=0 % constructors are not represented by partial applications
	 ; (getFuncArity(Name,Arity) -> Missing is Arity-TArity ; Missing=0)),
	(Missing=0 -> Term =.. [Name|Terms] ;
         (Missing>0 -> rev(Terms,RevTerms), Term = partcall(Missing,Name,RevTerms)
	             ; writeErr('INTERNAL COMPILER ERROR: over-application occured!'),
	               nlErr)),
	((CombType='FuncCall', Level=subterm)
	 -> NewShares = [makeShare(Term,V)|Shares], NewTerm = V
	  ; NewShares = Shares, NewTerm = Term).
exp2FuncShareTerm(Level,Vars,'Apply'(F,X),Shares,Term) :- !,
	exp2FuncShareTerm(subterm,Vars,F,FShares,SF),
	exp2FuncShareTerm(subterm,Vars,X,XShares,SX),
	append(FShares,XShares,FXShares),
	(Level=subterm -> Shares = [makeShare(apply(SF,SX),V)|FXShares], Term = V
	                ; Shares = FXShares, Term = apply(SF,SX)).
exp2FuncShareTerm(Level,Vars,'GuardedExpr'(LocalVars,Con,Exp),Shares,Term) :- !,
	append(Vars,LocalVars,NewVars),
	exp2FuncShareTerm(subterm,NewVars,Con,CShares,SCon),
	exp2FuncShareTerm(subterm,NewVars,Exp,EShares,SExp),
	append(CShares,EShares,CEShares),
	(Level=subterm -> Shares = [makeShare(cond(SCon,SExp),V)|CEShares], Term = V
	                ; Shares = CEShares, Term = cond(SCon,SExp)).
exp2FuncShareTerm(Level,Vars,'Constr'(LocalVars,Exp),Shares,Term) :- !,
	append(Vars,LocalVars,NewVars),
	exp2FuncShareTerm(Level,NewVars,Exp,Shares,Term).
exp2FuncShareTerm(_,_,Exp,[],'***unknown expression***') :-
	writeErr('INTERNAL ERROR in "exp2FuncShareTerm"!'),
	nlErr,
	writeErr('Unknown expression in FlatCurry file:'),
	nlErr,
	writeErr(Exp),
	nlErr,
	pleaseReport.

exp2FuncShareTerms(_,[],[],[]).
exp2FuncShareTerms(Vars,[T|Ts],Shares,[ST|STs]) :-
	exp2FuncShareTerm(subterm,Vars,T,Shares1,ST),
	exp2FuncShareTerms(Vars,Ts,Shares2,STs),
	append(Shares1,Shares2,Shares).

% add suffix to "makeShare":
addSuffix2MakeShare(Suffix,makeShare(V,NV),MSs) :-
	appendAtom(makeShare,Suffix,MakeShare),
	MSs =.. [MakeShare,V,NV].


% get the variables in a term which occurs more than once and must be shared:
getSharedVars(Exp,Shares) :-
	countVarsInTerm(Exp,[],Vars),
	varsIntoShares(Vars,Shares).

% get the variables of a case argument which occurs more tha once in the case expression
% and, thus, must be shared:
getSharedVarsOfCaseArg('Case'(_,CaseExp,Branches),Shares) :-
	countVarsInTerm(CaseExp,[],ArgVars),
	countVarsInBranches(Branches,ArgVars,CaseVars),
	prefixVars(ArgVars,CaseVars,MultArgVars),
	varsIntoShares(MultArgVars,Shares).

% prefixVars(Xs,Ys,Zs): Zs is prefix of Ys that has the same length as Xs
prefixVars([],_,[]).
prefixVars([_|Xs],[Y|Ys],[Y|Zs]) :- prefixVars(Xs,Ys,Zs).
	
varsIntoShares([],[]).
varsIntoShares([V/C|Vs],[makeShare(V,_)|Shs]) :-
	C>1, !,
	  % for statistics:
	  retract(numberOfShares(SC)),
	  SC1 is SC+1,
	  asserta(numberOfShares(SC1)),
	varsIntoShares(Vs,Shs).
varsIntoShares([_|Vs],Shs) :-
	varsIntoShares(Vs,Shs).

countVarsInTerm('Var'(V),Vars,IVars) :- !, incVarCount(V,Vars,IVars).
countVarsInTerm('Lit'(_),Vars,Vars) :- !.
countVarsInTerm('Comb'(_,_,Terms),Vars,IVars) :- !,
	countVarsInTerms(Terms,Vars,IVars).
countVarsInTerm('Apply'(F,X),Vars,IVars) :- !,
	countVarsInTerm(F,Vars,FVars),
	countVarsInTerm(X,FVars,IVars).
countVarsInTerm('GuardedExpr'(_,Con,Exp),Vars,IVars) :- !,
	countVarsInTerm(Con,Vars,CVars),
	countVarsInTerm(Exp,CVars,IVars).
countVarsInTerm('Constr'(_,Exp),Vars,IVars) :- !,
	countVarsInTerm(Exp,Vars,IVars).
countVarsInTerm('Choice'(Exp),Vars,IVars) :- !,
	countVarsInTerm(Exp,Vars,IVars).
countVarsInTerm('Case'(_,CaseExp,Branches),Vars,IVars) :- !,
	countVarsInTerm(CaseExp,Vars,CVars),
	countVarsInBranches(Branches,CVars,IVars).
countVarsInTerm('Or'(E1,E2),Vars,IVars) :- !,
	countVarsInTerm(E1,Vars,Vars1),
	countVarsInTerm(E2,Vars1,IVars).
countVarsInTerm(Exp,Vars,Vars) :- !,
	writeErr('INTERNAL ERROR in "countVarsInTerm"!'),
	nlErr,
	writeErr('Unknown expression in FlatCurry file:'),
	nlErr,
	writeErr(Exp),
	nlErr,
	pleaseReport.

countVarsInTerms([],Vars,Vars).
countVarsInTerms([X|Xs],Vars,IVars) :-
	countVarsInTerm(X,Vars,XVars),
	countVarsInTerms(Xs,XVars,IVars).

countVarsInBranches([],Vars,Vars).
countVarsInBranches(['Branch'(_,X)|Xs],Vars,IVars) :-
	countVarsInTerm(X,Vars,XVars),
	countVarsInBranches(Xs,XVars,IVars).

incVarCount(V,[],[V/1]).
incVarCount(V,[X/T|Vars],[X/T1|Vars]) :- V==X, !, T1 is T+1.
incVarCount(V,[X/T|Vars],[X/T|IVars]) :- incVarCount(V,Vars,IVars).



% select a variable in a list of variables.
% return position in the list and list of variables without this variable
selectVar(X,[Y|Ys],1,Ys) :- X==Y, !.
selectVar(X,[Y|Ys],XNr,[Y|Zs]) :- selectVar(X,Ys,Nr,Zs), XNr is Nr+1.



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Check correctness of FlatCurry rules and expressions.
% This checking is redundant and should only ensure
% that the front-end produces syntactically correct code.

check_flcFunction('Func'(Name,_,_,_,Rule)) :-
	flatName2Atom(Name,FName),
	retract(currentFunction(_)),
	asserta(currentFunction(FName)),
        check_flcRule(Rule), !.
check_flcFunction(_) :-
	writeErr('INTERNAL ERROR in FlatCurry file "'),
	writeErr('" in function "'),
	currentFunction(FName), writeErr(FName),
	writeErr('"!'), nlErr,
	pleaseReport.

check_flcRule('Rule'(Args,Expr)) :-
	map1(integer,Args),
	check_flcExpr(Expr).
check_flcRule('External'(Name)) :-
	check_flcString(Name).

check_flcString(S) :- map1(integer,S).

check_flcCaseType('Rigid').
check_flcCaseType('Flex').

check_flcCombType('FuncCall').
check_flcCombType('ConsCall').
check_flcCombType('PartCall').

check_flcExpr('Var'(_)) :- !.
check_flcExpr('Lit'(L)) :- !, check_flcLit(L).
check_flcExpr('Comb'(CT,Name,Es)) :- !,
	check_flcCombType(CT),
	check_flcString(Name),
	map1(check_flcExpr,Es).
check_flcExpr('Apply'(E1,E2)) :- !,
	check_flcExpr(E1), check_flcExpr(E2).
check_flcExpr('Constr'(Xs,E)) :- !,
	map1(integer,Xs),
	check_flcExpr(E).
check_flcExpr('Let'(Bindings,E)) :- !,
	map1(check_flcBinding,Bindings),
	check_flcExpr(E).
check_flcExpr('Or'(E1,E2)) :- !,
	check_flcExpr(E1), check_flcExpr(E2).
check_flcExpr('Case'(CT,E,Cs)) :- !,
	check_flcCaseType(CT),
	check_flcExpr(E),
	map1(check_flcCase,Cs).
check_flcExpr('GuardedExpr'(Xs,C,E)) :- !,
	map1(integer,Xs),
	check_flcExpr(C), check_flcExpr(E).
check_flcExpr('Choice'(E)) :- !,
	check_flcExpr(E).
check_flcExpr(E) :-
	writeErr('ERROR in FlatCurry file: Illegal expression "'),
	writeErr(E),
	writeErr('" in function "'),
	currentFunction(FName), writeErr(FName),
	writeErr('"!'), nlErr,
	reportLiftBug.

check_flcBinding('(,)'(V,E)) :- integer(V), check_flcExpr(E).

check_flcLit('Intc'(_)) :- !.
check_flcLit('Floatc'(_)) :- !.
check_flcLit('Charc'(_)) :- !.
check_flcLit(Lit) :-
	writeErr('ERROR in FlatCurry file: Illegal literal "'),
	writeErr(Lit),
	writeErr('" in function "'),
	currentFunction(FName), writeErr(FName),
	writeErr('"!'), nlErr,
	pleaseReport.

check_flcCase('Branch'('Pattern'(Name,Xs),E)) :- !,
	check_flcString(Name),
	map1(integer,Xs),
	check_flcExpr(E).
check_flcCase('Branch'('LPattern'(C),E)) :- !,
	check_flcLit(C),
	check_flcExpr(E).
check_flcCase(Branch) :-
	writeErr('ERROR in FlatCurry file: Illegal case pattern "'),
	writeErr(Branch),
	writeErr('" in function "'),
	currentFunction(FName), writeErr(FName),
	writeErr('"!'), nlErr,
	pleaseReport.


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Auxiliaries:

% generate N new free variables:
genVars(N,[]) :- N=0, !.
genVars(N,[_|Xs]) :- N1 is N-1, genVars(N1,Xs).

writeNTimes(0,_) :- !.
writeNTimes(N,S) :- N>0, write(S), N1 is N-1, writeNTimes(N1,S).

% write a clause and puts some initialization goals at the beginning
% of the right-hand side:
writeClauseWithInitGoals(LHS,Init,RHS) :-
	rev(Init,RInit),
	writeClauseWithRevInitGoals(LHS,RInit,RHS).
writeClauseWithRevInitGoals(LHS,[],RHS) :-
	writeClause((LHS :- RHS)).
writeClauseWithRevInitGoals(LHS,[G|Gs],RHS) :-
	writeClauseWithRevInitGoals(LHS,Gs,(G,RHS)).

writeClause(C) :-
	deleteLastTrue(C,C1),
	\+ \+ (numbervars(C1,0,_), writeq(C1)), write('.'), nl.

% delete a possibly generated last literal "true" in the clause:
deleteLastTrue((Head :- true),Head) :- !.
deleteLastTrue((Head :- B),(Head :- B1)) :- deleteLastTrueInBody(B,B1).
deleteLastTrue(Head,Head).

deleteLastTrueInBody((L,true),L) :- !.
deleteLastTrueInBody((L,B),(L,B1)) :- deleteLastTrueInBody(B,B1).


% check whether an atom consists only of alphanum chars:
isAlphaNumAtom(A) :- atom_chars(A,AS), isAlphaNumString(AS).
isAlphaNumString([]).
isAlphaNumString([C|Cs]) :-
	(65=<C, C=<90 ; C=95 ; 97=<C, C=<122),
	isAlphaNumString(Cs).

% replace all dots by underscores in an atom:
replaceDotByUnderscore(DA,UA) :-
	atom_chars(DA,DAS),
	replaceDotsByUnderscores(DAS,UAS),
	atom_chars(UA,UAS).
replaceDotsByUnderscores(DAS,UAS) :-
	append(S1,[46|S2],DAS), !,
	replaceDotsByUnderscores(S2,S3),
	append(S1,[95|S3],UAS).
replaceDotsByUnderscores(S,S).

% set compilation mode (i.e., fastcode: yes or no)
setFastCode(yes) :- prolog_flag(compiling,_,fastcode), !.
setFastCode(no) :- prolog_flag(compiling,_,compactcode), !.
setFastCode(_).

