%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Operations to load a translated Curry application.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- dynamic currentModule/1, loadedModule/1, importedModule/1, initializationsInProg/1,
	   currentCostCenters/1.

:- use_module(library(system)).
:- use_module(basics).
%:- use_module(compiler). % for generateMainPlFile in interactive mode

currentModule(''). % name of currently loaded module
initializationsInProg(true). % intializations to be done after loading Curry program
currentCostCenters(['']). % list of current cost centers


% set name of current Curry module:
curryModule(ModName) :-
	retract(currentModule(_)),
	asserta(currentModule(ModName)).

% set list of cost centers of the current Curry module:
costCenters(CCs) :-
	retract(currentCostCenters(_)),
	asserta(currentCostCenters(CCs)).

% initialize PAKCS before loading a new Curry application:
initializeBeforeLoad :-
	retractAllFacts(dynamicPredInfo/2),
	retractAllFacts(functiontype/6),
	retractAllFacts(constructortype/4),
	retractAllFacts(evaluation/2),
	retractAllFacts(loadedModule/1),
	retractAllFacts(importedModule/1),
	retract(initializationsInProg(_)),
	asserta(initializationsInProg(true)),
	retract(currentCostCenters(_)),
	asserta(currentCostCenters([''])).

% add initializations for currently loaded module:
initializationsInModule(ModInit) :-
	retract(initializationsInProg(ProgInit)),
	asserta(initializationsInProg((ProgInit,ModInit))).


% load a complete translated application by compiling the Prolog programs:
loadAndCompile(PrologFile,CreateMain) :-
	initializeBeforeLoad,
	compilePrologFile(PrologFile),
	currentModule(MainMod),
	assertz(loadedModule(MainMod)),
	loadAndCompileImports,
	curryModule(MainMod),
	(CreateMain=yes -> loadMain(PrologFile)
	                 ; appendAtom(PrologFile,'.main',MainPrologFile),
	                   compile(MainPrologFile)).

loadAndCompileImports :-
	retract(importedModule(Mod)),
	loadImportedModule(Mod),
	!,
	loadAndCompileImports.
loadAndCompileImports.

importModule(Mod) :- loadedModule(Mod), !. % already loaded
importModule(Mod) :- importedModule(Mod), !. % already in import set
importModule(Mod) :- assertz(importedModule(Mod)).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Auxiliaries for loading run-time libraries:

% directory containing the run-time modules:
moduleDir(MD) :-
        environ('PAKCSHOME',TCP),
        appendAtom(TCP,'/curry2prolog/',MD).

% ensure that run-time library is loaded:
ensure_lib_loaded(Lib) :- % first, look into working directory:
	working_directory(WDir,WDir),
	appendAtom(WDir,'/',Dir),
	appendAtom(Dir,Lib,DirLib),
	appendAtom(DirLib,'.pl',DirLibPl),
	file_exists(DirLibPl), !,
	ensure_loaded(DirLib).
ensure_lib_loaded(Lib) :-
	moduleDir(Dir),
	appendAtom(Dir,Lib,DirLib),
	ensure_loaded(DirLib).

loadImportedModule(Mod) :- loadedModule(Mod), !. % already loaded
loadImportedModule(Mod) :-
	appendAtom(Mod,'.pl',ModPl),
	(findFileInLoadPath(ModPl,PathModPl)
	 -> assertz(loadedModule(Mod)), compilePrologFile(PathModPl)
	  ; write(user_error,'ERROR: Compiled Curry module '),
	    write(user_error,ModPl),
	    write(user_error,' not found!'),
	    nl(user_error), !, fail).

compilePrologFile(PrologFileName) :-
	atom_chars(PrologFileName,PrologFileNameS),
	append(FileNameS,".pl",PrologFileNameS),
	append(FileNameS,".po",POFileNameS),
	atom_chars(POFileName,POFileNameS),
	(fileExistsAndNewer(POFileName,PrologFileName)
	 -> load_files(POFileName) % for faster compilation
	  ; compile(PrologFileName), try_save_program(PrologFileName)).

% start the CPNS demon (used in initialization of library Ports):
start_cpns_demon :-
	currentModule(cpns_demon),
	% don't automatically start the CPNS demon if the main program is
	% the CPNS demon itself (to avoid infinite recursion)
	!.
start_cpns_demon :-
	environ('PAKCSHOME',TCP),
        appendAtom(TCP,'/cpns/start',CPNSD),
	system(CPNSD).


