%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Interface to the CLP(FD) solver library of Sicstus-Prolog
%
% The clauses in this file are added to each compiled Curry program
% if finite domains constraints are used

:- use_module(library(clpfd)).

?- block prim_FD_domain(?,?,?,?,-,?).
prim_FD_domain(RL,RA,RB,R,E0,E) :-
        derefAll(RL,L), deref(RA,A), deref(RB,B), domain(L,A,B), R=success, E0=E.

?- block prim_FD_sum(?,?,?,?,-,?).
prim_FD_sum(RVs,RRel,RV,R,E0,E) :-
        derefAll(RVs,Vs),
	deref(RRel,RelCall), RelCall=partcall(2,FD_Rel,[]),
	translateFD_Rel(FD_Rel,Rel),
	deref(RV,V),
	sum(Vs,Rel,V), R=success, E0=E.

?- block prim_FD_scalar_product(?,?,?,?,?,-,?).
prim_FD_scalar_product(RCs,RVs,RRel,RV,R,E0,E) :-
        derefAll(RCs,Cs),
        derefAll(RVs,Vs),
	deref(RRel,RelCall), RelCall=partcall(2,FD_Rel,[]),
	translateFD_Rel(FD_Rel,Rel),
	deref(RV,V),
	scalar_product(Cs,Vs,Rel,V), R=success, E0=E.

?- block prim_FD_count(?,?,?,?,?,-,?).
prim_FD_count(RVal,RVs,RRel,RC,R,E0,E) :-
        deref(RVal,Val),
        derefAll(RVs,Vs),
	deref(RRel,RelCall), RelCall=partcall(2,FD_Rel,[]),
	translateFD_Rel(FD_Rel,Rel),
	deref(RC,C),
	count(Val,Vs,Rel,C), R=success, E0=E.

translateFD_Rel('CLPFD.=#',#=) :- !.
translateFD_Rel('CLPFD./=#',#\=) :- !.
translateFD_Rel('CLPFD.<#',#<) :- !.
translateFD_Rel('CLPFD.<=#',#=<) :- !.
translateFD_Rel('CLPFD.>#',#>) :- !.
translateFD_Rel('CLPFD.>=#',#>=) :- !.
translateFD_Rel(FD_Rel,_) :- writeErr('ERROR: Illegal FD constraint: '),
	writeErr(FD_Rel), nlErr, !, fail.

?- block prim_FD_all_different(?,?,-,?).
prim_FD_all_different(RL,R,E0,E) :-
        derefAll(RL,L), all_different(L), R=success, E0=E.

?- block prim_FD_indomain(?,?,-,?).
prim_FD_indomain(RV,R,E0,E) :-
	deref(RV,Var), indomain(Var), R=success, E0=E.

?- block prim_FD_labeling(?,?,?,-,?).
prim_FD_labeling(ROptions,RL,R,E0,E) :-
	derefAll(ROptions,Options),
	derefAll(RL,L),
	map2(translateLabelingOption,Options,PlOptions),
	labeling(PlOptions,L), R=success, E0=E.

translateLabelingOption('CLPFD.LeftMost',leftmost).
translateLabelingOption('CLPFD.FirstFail',ff).
translateLabelingOption('CLPFD.FirstFailConstrained',ffc).
translateLabelingOption('CLPFD.Min',min).
translateLabelingOption('CLPFD.Max',max).
translateLabelingOption('CLPFD.Step',step).
translateLabelingOption('CLPFD.Enum',enum).
translateLabelingOption('CLPFD.Bisect',bisect).
translateLabelingOption('CLPFD.Up',up).
translateLabelingOption('CLPFD.Down',down).
translateLabelingOption('CLPFD.All',all).
translateLabelingOption('CLPFD.Minimize'(DomVar),minimize(DomVar)).
translateLabelingOption('CLPFD.Maximize'(DomVar),maximize(DomVar)).
translateLabelingOption('CLPFD.Assumptions'(Var),assumptions(Var)).


?- block prim_FD_plus(?,?,?,-,?).
prim_FD_plus(RA,RB,R,E0,E) :-
	hnf(RA,A,E0,E1), hnf(RB,B,E1,E2), prim_FD_plus_exec(A,B,R,E2,E).

?- block prim_FD_plus_exec(?,?,?,-,?).
prim_FD_plus_exec(A,B,R,E0,E) :- R #= A+B, E0=E.

?- block prim_FD_minus(?,?,?,-,?).
prim_FD_minus(RA,RB,R,E0,E) :-
	hnf(RA,A,E0,E1), hnf(RB,B,E1,E2), prim_FD_minus_exec(A,B,R,E2,E).

?- block prim_FD_minus_exec(?,?,?,-,?).
prim_FD_minus_exec(A,B,R,E0,E) :- R #= A-B, E0=E.

?- block prim_FD_times(?,?,?,-,?).
prim_FD_times(RA,RB,R,E0,E) :-
	hnf(RA,A,E0,E1), hnf(RB,B,E1,E2), prim_FD_times_exec(A,B,R,E2,E).

?- block prim_FD_times_exec(?,?,?,-,?).
prim_FD_times_exec(A,B,R,E0,E) :- R #= A*B, E0=E.

?- block prim_FD_equal(?,?,?,-,?).
prim_FD_equal(RA,RB,R,E0,E) :-
	hnf(RA,A,E0,E1), hnf(RB,B,E1,E2), prim_constraint_exec(A #= B,R,E2,E).

?- block prim_FD_notequal(?,?,?,-,?).
prim_FD_notequal(RA,RB,R,E0,E) :-
	hnf(RA,A,E0,E1), hnf(RB,B,E1,E2), prim_constraint_exec(A #\= B,R,E2,E).

?- block prim_FD_le(?,?,?,-,?).
prim_FD_le(RA,RB,R,E0,E) :-
	hnf(RA,A,E0,E1), hnf(RB,B,E1,E2), prim_constraint_exec(A #< B,R,E2,E).

?- block prim_FD_leq(?,?,?,-,?).
prim_FD_leq(RA,RB,R,E0,E) :-
	hnf(RA,A,E0,E1), hnf(RB,B,E1,E2), prim_constraint_exec(A #=< B,R,E2,E).

?- block prim_FD_ge(?,?,?,-,?).
prim_FD_ge(RA,RB,R,E0,E) :-
	hnf(RA,A,E0,E1), hnf(RB,B,E1,E2), prim_constraint_exec(A #> B,R,E2,E).

?- block prim_FD_geq(?,?,?,-,?).
prim_FD_geq(RA,RB,R,E0,E) :-
	hnf(RA,A,E0,E1), hnf(RB,B,E1,E2), prim_constraint_exec(A #>= B,R,E2,E).

?- block prim_arithconstraint_exec(?,?,-,?).
prim_arithconstraint_exec(Exp,R,E0,E) :- R #= Exp, E0=E.

?- block prim_constraint_exec(?,?,-,?).
prim_constraint_exec(Con,success,E0,E) :- Con, E0=E.

