%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Definition of standard external functions:
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- use_module(library(system)).

% dereference a function's argument, i.e., remove all top-level sharing structures:
%deref(R,V) :- var(R), !, V=R.
%deref(share(M),V) :- !,
%	get_mutable(E,M), (E='$eval'(R) -> V=R ; deref(E,V)).
%deref(R,R).

% completely dereference a function's argument, i.e., remove all sharing structures
% also inside subterms:
%derefAll(R,V) :- var(R), !, V=R.
%derefAll(share(M),V) :- !,
%	get_mutable(E,M), (E='$eval'(R) -> derefAll(R,V) ; derefAll(E,V)).
%derefAll(R,V) :- R=..[C|As], map2(derefAll,As,Vs), V=..[C|Vs].

% evaluated an expression to constructor-rooted form and dereference result:
:- block constructorRootedAndDeref(?,?,-,?).
constructorRootedAndDeref(Exp,V,E0,E) :-
	hnf(Exp,Val,E0,E1), constructorRootedWaitAndDeref(Val,V,E1,E).
:- block constructorRootedWaitAndDeref(-,?,?,?),
	 constructorRootedWaitAndDeref(?,?,-,?).
constructorRootedWaitAndDeref(Val,V,E0,E) :-
	deref(Val,V), E0=E.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% definition of concurrent conjunction:
?- block prim_concurrent_and(?,?,?,-,?).
prim_concurrent_and(C1,C2,success,E0,E) :-
      hnf(C1,S1,E0,E1),
      hnf(C2,S2,E0,E2),
      waitconj(S1,S2,E1,E2,E).

?- block waitconj(?,?,-,?,?), waitconj(?,?,?,-,?).
waitconj(success,success,_,E,E).

% The always satisfiable primitive constraint:
prim_success(success,E,E).

% primitive for conditional rules:
?- block prim_cond(?,?,?,-,?).
prim_cond(Cond,RHS,R,E0,E) :- hnf(Cond,S,E0,E1), prim_checkcond(S,RHS,R,E1,E).
?- block prim_checkcond(-,?,?,?,?), prim_checkcond(?,?,?,-,?).
prim_checkcond(success,RHS,R,E0,E) :- hnf(RHS,R,E0,E).

% primitive for implementing letrec:
?- block prim_letrec(?,?,?,-,?).
prim_letrec(X,XE,success,E0,E) :- var(XE), !, X=XE, E0=E.
prim_letrec(X,XE,success,E0,E) :- create_mutable(XE,MX), X=share(MX), E0=E.


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Definition of arithmetic functions:
%
?- block prim_Int_plus(?,?,?,-,?).
prim_Int_plus(RA,RB,R,E0,E) :-
	hnf(RA,A,E0,E1), hnf(RB,B,E1,E2), prim_arithbinop_exec(A,B,A+B,R,E2,E).

?- block prim_Int_minus(?,?,?,-,?).
prim_Int_minus(RA,RB,R,E0,E) :-
	hnf(RA,A,E0,E1), hnf(RB,B,E1,E2), prim_arithbinop_exec(A,B,A-B,R,E2,E).

?- block prim_Int_times(?,?,?,-,?).
prim_Int_times(RA,RB,R,E0,E) :-
	hnf(RA,A,E0,E1), hnf(RB,B,E1,E2), prim_arithbinop_exec(A,B,A*B,R,E2,E).

?- block prim_Int_div(?,?,?,-,?).
prim_Int_div(RA,RB,R,E0,E) :-
	hnf(RA,A,E0,E1), hnf(RB,B,E1,E2), prim_arithbinop_exec(A,B,A//B,R,E2,E).

?- block prim_Int_mod(?,?,?,-,?).
prim_Int_mod(RA,RB,R,E0,E) :-
	hnf(RA,A,E0,E1), hnf(RB,B,E1,E2), prim_arithbinop_exec(A,B,A mod B,R,E2,E).

?- block prim_Int_le(?,?,?,-,?).
prim_Int_le(RA,RB,R,E0,E) :-
	hnf(RA,A,E0,E1), hnf(RB,B,E1,E2), prim_arithrel_exec(A,B,A<B,R,E2,E).

?- block prim_Int_ge(?,?,?,-,?).
prim_Int_ge(RA,RB,R,E0,E) :-
	hnf(RA,A,E0,E1), hnf(RB,B,E1,E2), prim_arithrel_exec(A,B,A>B,R,E2,E).

?- block prim_Int_leq(?,?,?,-,?).
prim_Int_leq(RA,RB,R,E0,E) :-
	hnf(RA,A,E0,E1), hnf(RB,B,E1,E2), prim_arithrel_exec(A,B,A=<B,R,E2,E).

?- block prim_Int_geq(?,?,?,-,?).
prim_Int_geq(RA,RB,R,E0,E) :-
	hnf(RA,A,E0,E1), hnf(RB,B,E1,E2), prim_arithrel_exec(A,B,A>=B,R,E2,E).

?- block prim_arithbinop_exec(-,?,?,?,?,?), prim_arithbinop_exec(?,-,?,?,?,?),
         prim_arithbinop_exec(?,?,?,?,-,?).
prim_arithbinop_exec(_,_,Exp,R,E0,E) :- R is Exp, E0=E.

?- block prim_arithunop_exec(-,?,?,?,?), prim_arithunop_exec(?,?,?,-,?).
prim_arithunop_exec(_,Exp,R,E0,E) :- R is Exp, E0=E.

?- block prim_arithrel_exec(-,?,?,?,?,?), prim_arithrel_exec(?,-,?,?,?,?),
         prim_arithrel_exec(?,?,?,?,-,?).
prim_arithrel_exec(_,_,Rel,R,E0,E) :- (Rel -> R='True' ; R='False'), E0=E.


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Definition of conversion functions for characters:
%
?- block prim_ord(?,?,-,?).
prim_ord(RC,N,E0,E) :- hnf(RC,C,E0,E1), prim_ord_exec(C,N,E1,E).

?- block prim_ord_exec(-,?,?,?), prim_ord_exec(?,?,-,?).
prim_ord_exec(RC,N,E0,E) :- deref(RC,C), char_int(C,N), E0=E.

?- block prim_chr(?,?,-,?).
prim_chr(RN,C,E0,E) :- hnf(RN,N,E0,E1), prim_chr_exec(N,C,E1,E).

?- block prim_chr_exec(-,?,?,?), prim_chr_exec(?,?,-,?).
prim_chr_exec(RN,C,E0,E) :- deref(RN,N), N>=0, N<256, !, char_int(C,N), E0=E.
prim_chr_exec(_,_,_,_) :-
	raise_exception('chr: argument out of range').


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Definition of I/O actions:
%
?- block prim_Monad_bind(?,?,?,?,-,?).
prim_Monad_bind(Action1,FunAction2,W,R,E0,E) :-
        prim_apply(Action1,W,'$io'(R1),E0,E1),
	hnf(FunAction2,HFunAction2,E1,E2),
	prim_apply(HFunAction2,R1,HAction2,E2,E3),
	prim_apply(HAction2,W,R,E3,E).

% although (>>) is a defined function,
% we provide a slightly faster primitive implementation:
?- block 'prim_>>'(?,?,?,?,-,?).
'prim_>>'(Action1,Action2,W,R,E0,E) :-
        prim_apply(Action1,W,_,E0,E1),
	hnf(Action2,HAction2,E1,E2),
	prim_apply(HAction2,W,R,E2,E).
?- block prim_Monad_seq(?,?,?,?,-,?).
prim_Monad_seq(Action1,Action2,W,R,E0,E) :-
        prim_apply(Action1,W,_,E0,E1),
	hnf(Action2,HAction2,E1,E2),
	prim_apply(HAction2,W,R,E2,E).

?- block prim_putChar(?,?,?,-,?).
prim_putChar(RC,_,R,E0,E) :- hnf(RC,C,E0,E1), prim_putChar_exec(C,R,E1,E).

?- block prim_putChar_exec(-,?,?,?), prim_putChar_exec(?,?,-,?).
prim_putChar_exec(C,'$io'('()'),E0,E) :-
	char_int(C,N), put(N), flush_output, E0=E.

?- block prim_getChar(?,?,-,?).
prim_getChar(_,'$io'(C),E0,E) :- get0(N), char_int(C,N), E0=E.

?- block prim_return(?,?,?,-,?).
prim_return(A,_,'$io'(A),E,E).

?- block prim_readFile(?,?,?,-,?).
prim_readFile(RA,_,_,_,_) :- derefAll(RA,A),
	map2(char_int,A,As),
	isURL(As), !,
	append("readFile """,As,E1),
	append(E1,""": URLs no longer supported in readFile!",E2),
	atom_chars(EMsg,E2),
	raise_exception(EMsg).
prim_readFile(RA,W,'$io'(Result),E0,E) :-
	derefAll(RA,A),
	string2Atom(A,FName),
	open(FName,read,Stream),
	(compileWithSharing(function)
	 -> makeShare(prim_readFileContents(Stream,W),Result)
	  ; Result = prim_readFileContents(Stream,W)),
	E0=E.

isURL(S) :- append("http://",_,S), !.
isURL(S) :- append("ftp://",_,S), !.

?- block 'prim_readFileContents'(?,?,?,-,?).
'prim_readFileContents'(Stream,_,Result,E0,E) :-
	at_end_of_stream(Stream), !,
	Result=[],
	close(Stream), E0=E.
'prim_readFileContents'(Stream,W,Result,E0,E) :-
	get0(Stream,NChar),
	char_int(Char,NChar),
	Result=[Char|RFC],
	(compileWithSharing(function)
	 -> makeShare(prim_readFileContents(Stream,W),RFC)
	  ; RFC = prim_readFileContents(Stream,W)),
	E0=E.

?- block prim_writeFile(?,?,?,?,-,?).
prim_writeFile(RA,S,W,H,E0,E) :- derefAll(RA,A),
	string2Atom(A,FName),
	open(FName,write,Stream),
	prim_writeFileContents(Stream,S,W,H,E0,E).	

?- block prim_appendFile(?,?,?,?,-,?).
prim_appendFile(RA,S,W,H,E0,E) :- derefAll(RA,A),
	string2Atom(A,FName),
	open(FName,append,Stream),
	prim_writeFileContents(Stream,S,W,H,E0,E).	

?- block prim_writeFileContents(?,?,?,?,-,?).
prim_writeFileContents(Stream,Contents,W,'$io'('()'),E0,E) :-
	hnf(Contents,HContents,E0,E1),
	prim_writeFileContents1(Stream,HContents,W,'$io'('()'),E1,E).
?- block prim_writeFileContents1(?,-,?,?,?,?),
         prim_writeFileContents1(?,?,?,?,-,?).
prim_writeFileContents1(Stream,[],_,'$io'('()'),E0,E) :-
	flush_output(Stream), close(Stream), E0=E.
prim_writeFileContents1(Stream,[C|Cs],W,R,E0,E) :-
	hnf(C,HC,E0,E1),
	eval_writeFileContents(Stream,HC,Cs,W,R,E1,E).
?- block eval_writeFileContents(?,-,?,?,?,-,?),
         eval_writeFileContents(?,?,?,?,?,-,?).
eval_writeFileContents(Stream,C,Cs,W,H,E0,E) :-
	char_int(C,N), put(Stream,N),
	prim_writeFileContents(Stream,Cs,W,H,E0,E).


?- block prim_catchFail(?,?,?,?,-,?).
prim_catchFail(Action,_,W,R,E0,E) :-
	on_exception(ErrorMsg,
		     hnf(apply(Action,W),R,E0,E),
		     (print_error(ErrorMsg), fail)),
	!.
prim_catchFail(_,ErrAction,W,R,E0,E) :-
	prim_apply(ErrAction,W,R,E0,E).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% definition of apply primitive:
?- block prim_apply(?,?,?,-,?).
prim_apply(F,X,R,E0,E) :- hnf(F,HF,E0,E1), prim_hnf_apply(HF,X,R,E1,E).

?- block prim_hnf_apply(-,?,?,?,?), prim_hnf_apply(?,?,?,-,?).
prim_hnf_apply(partcall(N,F,Args),X,R,E0,E) :- !,
	(N=1 -> prim_hnf_apply_call(F,Args,X,R,E0,E)
	      ; N1 is N-1, R=partcall(N1,F,[X|Args]), E0=E).
prim_hnf_apply(CTerm,X,R,E0,E) :- % partial constructor applications
	CTerm =.. [C|Args],
	append(Args,[X],ArgsX), % not constant time, hopefully occurs not so often,
	R =.. [C|ArgsX],        % otherwise one can represent them also as partcalls
	E0=E.

prim_hnf_apply_call(F,[],X,R,E0,E) :- !, Term =.. [F,X], hnf(Term,R,E0,E).
prim_hnf_apply_call(F,[A1],X,R,E0,E) :- !, Term =.. [F,A1,X], hnf(Term,R,E0,E).
prim_hnf_apply_call(F,[A1,A2],X,R,E0,E) :- !, Term =.. [F,A2,A1,X], hnf(Term,R,E0,E).
prim_hnf_apply_call(F,[A1,A2,A3],X,R,E0,E) :- !, Term =.. [F,A3,A2,A1,X], hnf(Term,R,E0,E).
prim_hnf_apply_call(F,Args,X,R,E0,E) :- % the general case:
	rev([X|Args],AllArgs), Term =.. [F|AllArgs], hnf(Term,R,E0,E).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% definition of seq primitive:
?- block prim_seq(?,?,?,-,?).
prim_seq(Arg,Val,Result,E0,E) :-
	hnf(Arg,_,E0,E1), hnf(Val,Result,E1,E).

% definition of ensureNotFree primitive (rigid on first argument):
?- block prim_ensureNotFree(?,?,-,?).
prim_ensureNotFree(Arg,Val,E0,E) :-
	hnf(Arg,Val,E0,E1), freeze(Val,E1=E).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% definition of runtime error:
?- block prim_error(?,?,-,?).
prim_error(RMsg,_,E,E) :-
	derefAll(RMsg,Msg),
	string2Atom(Msg,AMsg),
	raise_exception(AMsg).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% committed choice not yet implemented in Curry2Prolog:
?- block 'prim_commit'(?,?,-,?).
'prim_commit'(_,_,E,E) :-
        raise_exception('prelude.commit not yet implemented!').


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% encapsulated search not yet implemented in Curry2Prolog:
?- block 'prim_try'(?,?,-,?).
'prim_try'(_,_,E,E) :-
        raise_exception('prelude.try not yet implemented!').

?- block 'prim_tryone'(?,?,-,?).
'prim_tryone'(_,_,E,E) :-
        raise_exception('prelude.tryone not yet implemented!').


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% hack for implementing Curry's findall:
%
% Warning: in contrast to Curry's definition, this evaluation
% of findall is suspended until it contains no global variables
% (whereas according to Curry, findall is suspended only if
% it tries to bind a global variable).
% Moreover, it is strict, i.e., it evaluates always all solutions!

:- block prim_findall(?,?,-,?).
prim_findall(SG,Sols,E0,E) :- 
	waitUntilGround(SG,E0,E1), prim_findall_exec(SG,Sols,E1,E).

:- block prim_findall_exec(?,?,-,?).
prim_findall_exec(SG,Sols,E0,E) :- 
	hasPrintedFailure
	 -> findall(X,prim_apply(SG,X,success,E0,_),Sols), E0=E
	  ; asserta(hasPrintedFailure),
	    findall(X,prim_apply(SG,X,success,E0,_),Sols),
	    retract(hasPrintedFailure), E0=E.

:- block waitUntilGround(-,?,?), waitUntilGround(?,-,?).
waitUntilGround(share(M),E0,E) :-
	!,
	get_mutable(V,M),
	(V='$eval'(Exp) -> true ; Exp=V),
	waitUntilGround(Exp,E0,E).
waitUntilGround(T,E0,E) :- functor(T,_,N), waitUntilGroundArgs(1,N,T,E0,E).

:- block waitUntilGroundArgs(?,?,?,-,?).
waitUntilGroundArgs(A,N,_,E0,E) :- A>N, !, E0=E.
waitUntilGroundArgs(A,N,T,E0,E) :-
	arg(A,T,Arg), waitUntilGround(Arg,E0,E1),
	A1 is A+1, waitUntilGroundArgs(A1,N,T,E1,E).


% since the above implementation of findall is strict,
% we offer also findfirst which only evaluates the first solution:

:- block prim_findfirst(?,?,-,?).
prim_findfirst(SG,Sol,E0,E) :-
	waitUntilGround(SG,E0,E1), prim_findfirst_exec(SG,Sol,E1,E).

:- block prim_findfirst_exec(?,?,-,?).
prim_findfirst_exec(SG,Sol,E0,E) :-
	hasPrintedFailure
	 -> prim_findfirstWithPF(SG,Sol,E0), E0=E
	  ; asserta(hasPrintedFailure),
	    prim_findfirstWithoutPF(SG,Sol,E0), E0=E.

prim_findfirstWithPF(SG,Sol,E) :-
	prim_apply(SG,X,success,E,_), !, Sol=X.

prim_findfirstWithoutPF(SG,Sol,E) :-
	prim_apply(SG,X,success,E,_), retract(hasPrintedFailure), !, Sol=X.
prim_findfirstWithoutPF(_,_,_) :-
	retract(hasPrintedFailure), fail.


% We provide a similar implementation of AllSolutions.getFirstSolution:

:- block prim_getFirstSolution(?,?,-,?).
prim_getFirstSolution(RSG,_,Sol,E0,E) :-
	hnf(RSG,SG,E0,E1), waitUntilGround(SG,E1,E2),
	prim_getFirstSol_exec(SG,Sol,E2,E).

:- block prim_getFirstSol_exec(?,?,-,?).
prim_getFirstSol_exec(SG,Sol,E0,E) :-
	hasPrintedFailure
	 -> prim_getFirstSolWithPF(SG,Sol,E0,E)
	  ; asserta(hasPrintedFailure),
	    prim_getFirstSolWithoutPF(SG,Sol,E0,E).

prim_getFirstSolWithPF(SG,Sol,E0,E) :-
	prim_apply(SG,X,success,E0,E1), !, Sol='$io'('Just'(X)), E1=E.
prim_getFirstSolWithPF(_,'$io'('Nothing'),E,E).

prim_getFirstSolWithoutPF(SG,Sol,E0,E) :-
	prim_apply(SG,X,success,E0,E1), retract(hasPrintedFailure), !,
	Sol='$io'('Just'(X)), E1=E.
prim_getFirstSolWithoutPF(_,'$io'('Nothing'),E0,E) :-
	retract(hasPrintedFailure), E0=E.


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Directed non-strict equality for matching against functional patterns:
% (first argument must be the functional pattern):
:-block unifEq(?,?,?,-,?).
unifEq(A,B,success,E0,E):- hnf(A,HA,E0,E1), unifEq1(HA,B,E1,E).

:-block unifEq1(?,?,-,?).
% In the following clause, we bind a functional pattern variable to the actual argument.
% This binding of a logical variable against a non-constructor term is not problematic
% since the functional pattern variable is a logical variable that is not enclosed
% by a sharing structure (compare definition of makeShare).
unifEq1(A,B,E0,E):- var(A), !, occursNot(A,B), A=B,
	%writeErr('BOUND TO: '), removeShares(B,UB), writeErr(UB), nlErr,
	E0=E.
unifEq1(A,B,E0,E):-
	replaceMultipleVariables(A,LinA,LinConstraints),
	hnf(B,HB,E0,E1),
	unifEqHnf(LinA,HB,E1,E2),
	%(LinConstraints=success -> true
	% ; writeErr('Linearity constraints: '), writeErr(LinConstraints), nlErr),
	hnf(LinConstraints,_,E2,E).

% replace multiple occurrences of a same logic variables by new ones combined
% with strict equations:
replaceMultipleVariables(T,LinT,LinConstraints) :-
	%writeErr('Term to linearize: '), writeErr(T), nlErr,
	T =.. [Cons|Args],
	replaceMultipleVariablesInArgs(Args,inConstructorCall,Vars,LinArgs),
	LinT =.. [Cons|LinArgs],
	getSEqConstraints(Vars,LinConstraints).
	%length(Vars,Len), writeErr('Length of variable list: '), writeErr(Len), nlErr.

getControlVar(X,Below,L,NewX) :- var(L), !, L=[control(X,Below,NewX,_)|_].
getControlVar(X,Below,[control(Y,YBelow,NewVar,NewConstraint)|_],NewX) :- X==Y, !,
	% multiple occurrence of a variable X not inside function calls are replaced
	% a fresh variable Y and a strict equality constraint X=:=Y
	% that is later executed.
	((Below=inConstructorCall, YBelow=inConstructorCall)
	 -> (var(NewConstraint)
	     -> NewConstraint = '=:='(X,NewVar), NewX=X
 	      ; NewX=NewVar)
	  ; % multiple occurrence of a variable X, where one occurrence is in a
	    % function call, are replaced by an expression that
	    % forces the evaluation of a strict equality constraint if the variable
	    % occurs multiple times in the finally evaluated pattern.
	    % Therefore, each variable X is replaced by
	    % (if isVar ShareX then ShareX=:=() else CtrlX=:=()) &> X
	    % with a constraint
	    % (if isVar CtrlX then success else X=:=X) that is later executed.
	    NewX=NewVar,
	    (var(NewConstraint)
	     -> NewVar = '&>'(ifVar(ShareVar,'=:='(ShareVar,'()'),'=:='(CtrlVar,'()')),X),
	        NewConstraint = ifVar(CtrlVar,success,'=:='(X,X))
 	      ; true)).
getControlVar(X,Below,[_|L],NewVar) :- getControlVar(X,Below,L,NewVar).

getSEqConstraints(L,success) :- var(L), !, L=[].
getSEqConstraints([control(X,_,NewVar,NewConstraint)|L],Constraints) :-
	var(NewConstraint), !, % occurred only once
	X=NewVar,
	getSEqConstraints(L,Constraints).
getSEqConstraints([control(_,_,_,NewConstraints)|L],'&'(NewConstraints,Constraints)) :-
	getSEqConstraints(L,Constraints).

replaceMultipleVariablesInArgs([],_,_,[]).
replaceMultipleVariablesInArgs([X|Args],Below,Vars,[NewArg|LinArgs]) :-
	var(X), !, getControlVar(X,Below,Vars,NewArg),
	replaceMultipleVariablesInArgs(Args,Below,Vars,LinArgs).
replaceMultipleVariablesInArgs([Arg|Args],Below,Vars,[LinArg|LinArgs]) :-
	Arg =.. [FC|Ts],
	(functiontype(FC,_,_,_,_,_) -> TsBelow= inFunctionCall ; TsBelow=Below),
	replaceMultipleVariablesInArgs(Ts,TsBelow,Vars,LinTs),
   	LinArg =.. [FC|LinTs],
	replaceMultipleVariablesInArgs(Args,Below,Vars,LinArgs).

:-block unifEqHnf(?,?,-,?).
unifEqHnf(A,B,E0,E):-var(B),!, bind(B,A,E0,E). % in order to evaluate function pattern
unifEqHnf(A,B,E0,E):-integer(A),!,A=B,E0=E.
unifEqHnf(A,B,E0,E):-float(A),!,{A=B},E0=E.
unifEqHnf(A,B,E0,E) :-
	functor(A,FuncA,ArA), functor(B,FuncB,ArB), FuncA==FuncB, ArA==ArB,
	genUnifEqHnfBody(1,ArA,A,B,Con),hnf(Con,_,E0,E).

genUnifEqHnfBody(N,Arity,_,_,success) :- N>Arity, !.
genUnifEqHnfBody(N,Arity,A,B,'=:<='(ArgA,ArgB)):-
	N=Arity, !,
	arg(N,A,ArgA), arg(N,B,ArgB).
genUnifEqHnfBody(N,Arity,A,B,&('=:<='(ArgA,ArgB),G)):-
	arg(N,A,ArgA), arg(N,B,ArgB),
	N1 is N+1,
	genUnifEqHnfBody(N1,Arity,A,B,G).

% Directed non-strict equality for matching against linear functional patterns,
% i.e., it must be ensured that the first argument is always (after evalutation
% by narrowing) a linear pattern.
% At call time, the first argument must be the functional pattern.
:-block unifEqLinear(?,?,?,-,?).
unifEqLinear(A,B,success,E0,E):- hnf(A,HA,E0,E1), unifEqLinear1(HA,B,E1,E).

:-block unifEqLinear1(?,?,-,?).
% In the following clause, we bind a functional pattern variable to the actual argument.
% This binding of a logical variable against a non-constructor term is not problematic
% since the functional pattern variable is a logical variable that is not enclosed
% by a sharing structure (compare definition of makeShare).
unifEqLinear1(A,B,E0,E):- var(A), !, A=B,
	%writeErr('BOUND TO: '), removeShares(B,UB), writeErr(UB), nlErr,
	E0=E.
unifEqLinear1(A,B,E0,E):- hnf(B,HB,E0,E1), unifEqLinearHnf(A,HB,E1,E).

:-block unifEqLinearHnf(?,?,-,?).
unifEqLinearHnf(A,B,E0,E) :- var(B), !,
	nf(A,NA,E0,E1), B=NA, E1=E.
	%bind(B,A,E0,E).
unifEqLinearHnf(A,B,E0,E) :- integer(A), !, A=B, E0=E.
unifEqLinearHnf(A,B,E0,E) :- float(A), !, {A=B}, E0=E.
unifEqLinearHnf(A,B,E0,E) :-
	functor(A,FuncA,ArA), functor(B,FuncB,ArB), FuncA==FuncB, ArA==ArB,
	genUnifEqLinearHnfBody(1,ArA,A,B,Con),hnf(Con,_,E0,E).

genUnifEqLinearHnfBody(N,Arity,_,_,success) :- N>Arity, !.
genUnifEqLinearHnfBody(N,Arity,A,B,'=:<<='(ArgA,ArgB)):-
	N=Arity, !,
	arg(N,A,ArgA), arg(N,B,ArgB).
genUnifEqLinearHnfBody(N,Arity,A,B,&('=:<<='(ArgA,ArgB),G)):-
	arg(N,A,ArgA), arg(N,B,ArgB),
	N1 is N+1,
	genUnifEqLinearHnfBody(N1,Arity,A,B,G).


% ifVar x t f  corresponds to  if (Unsafe.isVar x) then t else f:
?- block prim_ifVar(?,?,?,?,-,?).
prim_ifVar(RTerm,T,F,H,E0,E) :-
	deref(RTerm,Term),
        (var(Term) -> hnf(T,H,E0,E) ; hnf(F,H,E0,E)).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
