---------------------------------------------------------------------
-- GUI for CurryBrowser, a generic analysis environment for declarative
-- programs.
---------------------------------------------------------------------

module BrowserGUI where

import GUI
import Read
import IOExts
import Imports
import FlatCurry
import FlatCurryGoodies
import FlatCurryTools
import ShowFlatCurry(leqFunc)
import System
import FileGoodies
import List
import Maybe
import AnalysisTypes
import BrowserAnalysis
import Sort
import Dependency(callsDirectly,indirectlyDependent)
import ShowGraph
import ImportCalls

---------------------------------------------------------------------
-- Set this constant to True if the execution times of the main operations
-- should be shown in the status line:
showExecTime = False

---------------------------------------------------------------------
-- Title and version
title = "Curry Browser"

version = "Version of 30/09/2005"

patchReadmeVersion = do
  readmetxt <- readCompleteFile "README"
  writeFile "README" (unlines . map replaceVersion . lines $ readmetxt)
 where
  replaceVersion l = if isPrefixOf "Version" l then version else l

---------------------------------------------------------------------
-- Main program: check arguments, read interfaces, and run GUI:
main = do
  args <- getArgs
  case args of
   [a] -> start (baseName (stripSuffix a))
   _ -> putStrLn $ "ERROR: Illegal arguments for Curry Browser: " ++
                   concat (intersperse " " args) ++ "\n" ++
                   "Usage: currybrowser <module_name>"

start mod =
  putStrLn "Please be patient, reading all interfaces..." >>
  getBrowserDir >>= \dir ->
  readFile (dir++"README") >>= \helptxt ->
  getImportedInterfaces mod >>= \mods ->
  let mainmod = moduleName (progOfIFFP (snd (head mods)))
      trees = [Leaf (if mainmod==mod then mod else mod++"(module: "++mainmod++")")
                    (mainmod,map (moduleImports . progOfIFFP . snd) mods)]
   in newIORef (trees,
                (mainmod,(mod,snd (head mods))):map (\(m,p)->(m,(m,p))) (tail mods),
                [],("",OtherText,""),[]) >>= \stateref ->
      runInitGUI (title++" ("++version++")")
                 (browserGUI stateref rmod rtxt (trees2strings trees))
                 (\gp -> setValue rtxt helptxt gp >> setValue rmod "0" gp)
 where rmod,rtxt free

m1 = start "BrowserGUI"
m2 = start "Imports"
m3 = start "demo"
m4 = start "rev"

-----------------------------------------------------------------------------
-- Structure of import tree to be shown:
data Tree a = Leaf String a          -- leaves have a name and a value
            | Node String a [Tree a] -- nodes have a name, value and a list of subtrees

trees2strings trees = concatMap (tree2strings 0) trees

tree2strings n (Leaf t _)       = [blanks n ++ "+ "++t]
tree2strings n (Node t _ trees) = (blanks n ++"- "++t)
                                  : concatMap (tree2strings (n+2)) trees

blanks n = take n (repeat ' ')


-- get name of selected node in the tree list:
getTreesNodeName :: Int -> [Tree _] -> String
getTreesNodeName _ [] = error "getTreesNodeName: nothing selected" -- should not occur
getTreesNodeName n (Leaf name _ : trees) =
  if n==0 then name
          else getTreesNodeName (n-1) trees
getTreesNodeName n (Node t v subtrees : trees) =
  if n==0 then t
          else let l = length (tree2strings 0 (Node t v subtrees)) in
               if n < l
               then getTreesNodeName (n-1) subtrees
               else getTreesNodeName (n-l) trees

-- get selected value in the tree list:
getTreesValue :: Int -> [Tree a] -> a
getTreesValue _ [] = error "getTreesValue: nothing selected" -- should not occur
getTreesValue n (Leaf _ v : trees) =
  if n==0 then v 
          else getTreesValue (n-1) trees
getTreesValue n (Node t v subtrees : trees) =
  if n==0 then v
          else let l = length (tree2strings 0 (Node t v subtrees)) in
               if n < l
               then getTreesValue (n-1) subtrees
               else getTreesValue (n-l) trees

-- change tree when click on some line (i.e., open or close a node):
changeTrees _ [] = return [] -- should not occur
changeTrees n (Leaf t v : trees) =
  if n==0 then openNode v >>= \newst -> return (Node t v newst : trees)
          else changeTrees (n-1) trees >>= \nts -> return (Leaf t v : nts)
changeTrees n (Node t v subtrees : trees) =
  if n==0 then return (Leaf t v : trees)
          else let l = length (tree2strings 0 (Node t v subtrees)) in
               if n < l
               then (changeTrees (n-1) subtrees) >>= \nsts ->
                    return (Node t v nsts : trees)
               else changeTrees (n-l) trees >>= \nts -> return (Node t v subtrees : nts)

openNode (mod,modimps) = let mbimps = lookup mod modimps in
  return $  maybe [] (map (\m->Leaf m (m,modimps))) mbimps

-----------------------------------------------------------------------------
-- Operations on the state of the GUI:
-- The GUI state consists of
-- * the import tree
-- * the list of all modules (module name, file name, FlatCurry interface or program)
-- * the list of all currently shown functions of the module
-- * the name and contents of the module currently shown in the contents window
-- * the list of qualified function names shown in the function search widget

type ImportTree = [Tree (String,[(String,[String])])]


type GuiState = (ImportTree,[(String,(String,InterfaceOrFlatProg))],
                 [FuncDecl],(String,ContentsKind,String),[QName])

getModuleFileName gs mod =
    readIORef gs >>= \(_,mods,_,_,_) -> return (fst (fromJust (lookup mod mods)))


getTrees gs = readIORef gs >>= \(trees,_,_,_,_) -> return trees

storeTrees gs trees =
   readIORef gs >>= \(_,ms,_,ct,fns) -> writeIORef gs (trees,ms,[],ct,fns)

-- extract all reflexive-transitive imports for a module from GUI state:
getAllImportsOfModule gs mod = do
  (trees,_,_,_,_) <- readIORef gs
  return (collectImports (importsOfRoot trees) [mod] [])
 where
   importsOfRoot ((Leaf _ (_,imps)) :_)   = imps
   importsOfRoot ((Node _ (_,imps) _) :_) = imps

   collectImports _ [] imps = imps
   collectImports modimps (m:ms) doneimps =
    if m `elem` doneimps
    then collectImports modimps ms doneimps
    else collectImports modimps (ms ++ fromJust (lookup m modimps)) (m:doneimps)

getFuns gs = readIORef gs >>= \(_,_,funs,_,_) -> return funs

storeModuleFunctions gs (Prog _ _ _ funs _) = do
  (t,ms,_,ct,fns) <- readIORef gs
  writeIORef gs (t,ms,mergeSort leqFunc funs,ct,fns)

storeExportedFunctions gs (Prog _ _ _ funs _) = do
  (t,ms,_,ct,fns) <- readIORef gs
  writeIORef gs
     (t,ms,mergeSort leqFunc (filter (\(Func _ _ vis _ _)->vis==Public) funs),ct,fns)

setMainContentsModule :: IORef GuiState -> String -> ContentsKind -> String -> IO ()
setMainContentsModule gs cntmod cntkind contents = do
  (t,ms,fs,_,fns) <- readIORef gs
  writeIORef gs (t,ms,fs,
                 (if cntkind==OtherText then "" else cntmod,cntkind,contents),fns)

getContentsModule :: IORef GuiState -> IO String
getContentsModule gs = do
  (_,_,_,(cntmod,_,_),_) <- readIORef gs
  return cntmod

getMainContents :: IORef GuiState -> IO (ContentsKind,String)
getMainContents gs = do
  (_,_,_,(_,cntkind,contents),_) <- readIORef gs
  return (cntkind,contents)

getFindFunctions :: IORef GuiState -> IO [QName]
getFindFunctions gs = do
  (_,_,_,_,fns) <- readIORef gs
  return fns

setFindFunctions :: IORef GuiState -> [QName] -> IO ()
setFindFunctions gs fns = do
  (t,ms,fs,cnttype,_) <- readIORef gs
  writeIORef gs (t,ms,fs,cnttype,fns)

-- get the interfaces (or FlatCurry programs) of all modules:
getAllModules gs = do
  (_,mods,_,_,_) <- readIORef gs
  return (map (progOfIFFP . snd . snd) mods)

-- get the interface (or FlatCurry program) of a specific module:
getIntWithName gs name = do
  (_,mods,_,_,_) <- readIORef gs
  return (progOfIFFP . snd . fromJust . lookup name $ mods)

-- get the FlatCurry program of a specific module (read, if necessary):
getProgWithName gs prt name = do
  (_,mods,_,_,_) <- readIORef gs
  ifOrProg (\_->readProgAndStore gs prt name) return (snd (fromJust (lookup name mods)))

-- Get all data type declarations (also imported) of a module:
getAllTypes gs _ mod =
  getAllImportsOfModule gs mod >>= \imps ->
  readIORef gs >>= \(_,mods,_,_,_) ->
  return (concatMap (typesOfProg . progOfIFFP . snd . snd)
                    (filter ((`elem` imps) . fst) mods))

-- Get all function declarations (also imported) of a module:
getAllFunctions gs prt mod =
  getAllImportsOfModule gs mod >>= \imps ->
  readIORef gs >>= \(_,mods,_,_,_) ->
  mapIO_ (readProgAndStoreIfNecessary gs prt) (filter ((`elem` imps) . fst) mods) >>
  readIORef gs >>= \(_,newmods,_,_,_) ->
  return (concatMap (funcsOfProg . progOfIFFP . snd. snd)
                    (filter ((`elem` imps) . fst) newmods))

-- Get all function names (also imported) of a module:
getAllFunctionNames :: IORef GuiState -> String -> IO [QName]
getAllFunctionNames gs mod =
  getAllImportsOfModule gs mod >>= \imps ->
  readIORef gs >>= \(_,mods,_,_,_) ->
  return (map funcName (concatMap (funcsOfProg . progOfIFFP . snd. snd)
                                  (filter ((`elem` imps) . fst) mods)))

-- read a FlatCurry program and store if not already done
readProgAndStore gs prt mod = do
  (ts,mods,fs,ct,fns) <- readIORef gs
  prog <- readFlatCurryFileInLoadPath prt (fst (fromJust (lookup mod mods)))
  writeIORef gs (ts,update mod prog mods,fs,ct,fns)
  return prog
 where
  update _ _ [] = []
  update nm pr ((n,(f,p)):ms) = if n==nm then (n,(f,FP pr)):ms
                                         else (n,(f,p)) : update nm pr ms

-- read a FlatCurry program and store if not already done
readProgAndStoreIfNecessary _ _ (_,(_,FP _)) = done
readProgAndStoreIfNecessary gs prt (name,(_,IF _)) =
   readProgAndStore gs prt name >> done

-----------------------------------------------------------------------------
-- The GUI of the browser:

browserGUI :: IORef GuiState -> WidgetRef -> WidgetRef -> [String] -> Widget
browserGUI gstate rmod rtxt names =
 col [
   row [
     Col [LeftAlign] [
       Label [Text "Select module and imports:"],
       ListBoxScroll [WRef rmod, List names, Width 20, Height 14,
                      Cmd (showBusy selmod), Background "yellow", Fill],
       row [Label [Text "Functions:"],
               Button (showBusy (executeForModule showExportedFuns)) [Text "Exported"],
               Button (showBusy (executeForModule showAllModuleFuns)) [Text "All"]],
       ListBoxScroll [WRef rfun, Width 20, Height 16,
                      Cmd (showBusy selfun), Background "white", Fill]],
     Col [LeftAlign] [
       row [MenuButton
             [Text "Select...",
              Menu [MButton (showMBusy (executeForModule showAllModuleFunctionsToFocus))
                            "all functions in selected module",
                    MButton (showMBusy (executeForModule showAllFunctionsToFocus))
                            "all exported functions in selected and imported modules",
                    MButton (showMBusy showDirectCalls)
                            "all direct dependants from selected function",
                    MButton (showMBusy showInDirectCalls)
                            "all dependants from selected function"]],
             Label [Text "to focus in code:"]],
       ListBoxScroll [WRef rfindfuns, Width 20,
                      Height (32 - length functionAnalyses - length allFunctionAnalyses),
                      Cmd (showBusy focusFunction), Background "white", Fill],
       Label [Text "Analyse all functions:"],
       ListBox [WRef ranaall, Width 20, Height (length allFunctionAnalyses),
                List (map (\(n,_,_)->n) allFunctionAnalyses),
                Cmd (showBusy analyzeAllFuns), Background "white", Fill],
       Label [Text "Property of selected function:"],
       ListBox [WRef rana, Width 20, Height (length functionAnalyses),
                List (map fst functionAnalyses),
                Cmd (showBusy selfun), Background "white", Fill]],
     Col [LeftAlign] [
       row [Button (showBusy (executeForModule showSource)) [Text "Show source"],
            MenuButton
              [Text "Show selected module as...",
               Menu (map (\ (t,ma) -> MButton (showMBusy (executeForModule
                                                           (analyseModuleWith ma))) t)
                            moduleAnalyses)],
            MenuButton
              [Text "Tools...",
               Menu [MButton (showMBusy (executeForModule showImpCalls))
                             "List calls to imported functions in selected module",
                     MButton (showMBusy showImportGraph)
                             "Show import graph of all modules (except prelude) (DOT)"]],
            MenuButton
              [Text "File...",
               Menu [MButton (showMBusy saveMainText) "Save program text as...",
                     MSeparator,
                     MButton (\gp->exitGUI gp >> return []) "Exit"]],
            MenuButton
              [Text "Help...",
               Menu [MButton (showMBusy (help "README")) "About the Curry Browser",
                     MSeparator,
                     MButton (showMBusy (help "Help.txt"))
                             "How to use the Curry Browser",
                     MButton (showMBusy (help "Extend.txt"))
                             "How to extend the Curry Browser"]]
              ],
       TextEditScroll [WRef rtxt, Height 25, Width 72, Background "white"],
       Label [Text "Comments and results:"],
       TextEditScroll [WRef resultwidget, Height 5, Width 72, Background "white"]]],
   Label [WRef rstatus, Text "Status: ready", Background "green", FillX]]
 where
  rfindfuns,resultwidget,rfun,rana,ranaall,rstatus free

  saveMainText gp = do
    file <- getSaveFile
    if null file
     then done
     else getValue rtxt gp >>= writeFile file

  -- put a message in main contents widget:
  putMainMessage gp msg = do
    setValue rtxt msg gp
    setMainContentsModule gstate "" OtherText msg

  -- show info texts:
  help localhelpfile gp =
    getBrowserDir >>= \dir ->
    readFile (dir++localhelpfile) >>=
    putMainMessage gp

  -- show business while executing an event handler:
  showBusy handler gp = do
     setConfig rstatus (Background "red") gp
     setConfig rstatus (Text "Status: running") gp
     time1 <- getCPUTime
     handler gp
     time2 <- getCPUTime
     setConfig rstatus (Text $ if showExecTime
                               then "Status: ready (" ++ show(time2-time1) ++ " msecs)"
                               else "Status: ready") gp
     setConfig rstatus (Background "green") gp

  showMBusy handler gp = showBusy handler gp >> return []

  -- show what we are doing in status line:
  showDoing gp str = setConfig rstatus (Text ("Status: "++str)) gp

  -- Execute an I/O action safely, i.e., catch all errors and failures:
  safeIO gp act = catchFail act (putMainMessage gp "Failure occurred!")

  -- click on a module name in left module column:
  selmod gp =
    getValue rmod gp >>= \sel ->
    if sel==""
      then done
      else putMainMessage gp "" >>
           setConfig rfun (List []) gp >>
           getTrees gstate >>= \trees ->
           changeTrees (readNat sel) trees >>= \newtrees ->
           storeTrees gstate newtrees >>
           setConfig rmod (List (trees2strings newtrees)) gp >>
           setValue resultwidget "" gp >>
           setValue rmod sel gp

  -- get the name of the selected module (or Nothing in case of no selection):
  getSelectedModName gp = do
    sel <- getValue rmod gp
    if sel==""
      then return Nothing
      else getTrees gstate >>= \trees ->
           return (Just (fst (getTreesValue (readNat sel) trees)))

  -- execute event handler on the selected module (or show "nothing selected" message):
  executeForModule modhandler gp =
    getSelectedModName gp >>= \mod ->
    if mod==Nothing
      then putMainMessage gp "No module selected!"
      else modhandler (fromJust mod) gp

  -- analyse a selected module:
  analyseModuleWith modanalysis mod gp = safeIO gp $
    performModuleAnalysis modanalysis (showDoing gp) mod >>= \res ->
    showModAnalysisResult mod res gp
    
  showModAnalysisResult mod (ContentsResult cntkind contents) gp = do
    setValue rtxt contents gp
    setMainContentsModule gstate mod cntkind contents
  showModAnalysisResult _ (ModuleAction act) _ = act

  -- show module source code:
  showSource mod gp = do
    fmod <- getModuleFileName gstate mod
    mbprogname <- Imports.findFileInLoadPath fmod [".lcurry",".curry"]
    maybe (putMainMessage gp "Source file does not exist!")
          (\filename -> readFile filename >>= \source ->
                        setValue rtxt source gp >>
                        setMainContentsModule gstate mod
                            (if take 7 (reverse filename) == "yrrucl."
                             then LCurryProg else CurryProg) source)
          mbprogname

  -- show module dependency graph:
  showImportGraph gp =
    getAllModules gstate >>= \mods ->
    safeIO gp $
       viewDependencyGraph
         (concatMap (\(Prog mod imps _ _ _) ->
                          if mod=="prelude" then []
                                            else [(mod,[],delete "prelude" imps)]) mods)

  -- show import calls of selected module:
  showImpCalls mod gp =
    getProgWithName gstate (showDoing gp) mod >>= \prog ->
    putMainMessage gp (showImportCalls prog)

  -- show module's functions:
  showAllModuleFuns mod gp = do
    prog <- getProgWithName gstate (showDoing gp) mod
    storeModuleFunctions gstate prog
    funs <- getFuns gstate
    setConfig rfun (List (map unqualifiedName funs)) gp

  -- show module's exported functions:
  showExportedFuns mod gp = do
    prog <- getProgWithName gstate (showDoing gp) mod
    storeExportedFunctions gstate prog
    funs <- getFuns gstate
    setConfig rfun (List (map unqualifiedName funs)) gp

  -- click on a module name in function column:
  selfun gp = safeIO gp $ 
    getSelectedModName gp >>= \mod ->
    getValue rfun gp >>= \self ->
    getValue rana gp >>= \sela ->
    if mod==Nothing || null self || null sela then done else
      getFuns gstate >>= \funs ->
      performAnalysis (snd (functionAnalyses!!readNat sela)) (showDoing gp)
                      (fromJust mod) (funs!!(readNat self)) >>= \anaresult ->
      showAnalysisResult anaresult gp

  showAnalysisResult (MsgResult str) gp = setValue resultwidget str gp
  showAnalysisResult (ActionResult act) _ = act

  -- show directly dependent functions to focus:
  showDirectCalls gp =
    getSelectedModName gp >>= \mod ->
    getValue rfun gp >>= \self ->
    if mod==Nothing || null self then setConfig rfindfuns (List []) gp else
      getFuns gstate >>= \funs ->
      let mainfun = funs!!(readNat self)
          qfnames = mergeSort leqQName (union [funcName mainfun] (callsDirectly mainfun))
       in setFindFunctions gstate qfnames >>
          setConfig rfindfuns (List (map (\(m,n)->n++" ("++m++")") qfnames)) gp

  -- show indirectly dependent functions to focus:
  showInDirectCalls gp =
    getSelectedModName gp >>= \mod ->
    getValue rfun gp >>= \self ->
    if mod==Nothing || null self then setConfig rfindfuns (List []) gp else
      getFuns gstate >>= \funs ->
      let mainfun = funcName (funs!!(readNat self)) in
      getAllFunctions gstate (showDoing gp) (fromJust mod) >>= \allfuns ->
      let qfnames = mergeSort leqQName
             (union [mainfun] (fromJust (lookup mainfun (indirectlyDependent allfuns))))
       in setFindFunctions gstate qfnames >>
          setConfig rfindfuns (List (map (\(m,n)->n++" ("++m++")") qfnames)) gp

  -- show indirectly dependent functions to focus:
  showAllModuleFunctionsToFocus mod gp =
    getProgWithName gstate (showDoing gp) mod >>= \prog ->
    let qfnames = mergeSort leqQName (map funcName (funcsOfProg prog))
     in setFindFunctions gstate qfnames >>
        setConfig rfindfuns (List (map (\(m,n)->n++" ("++m++")") qfnames)) gp

  -- show indirectly dependent functions to focus:
  showAllFunctionsToFocus mod gp =
    getAllFunctionNames gstate mod >>= \allfuns ->
    let qfnames = mergeSort leqQName allfuns
     in setFindFunctions gstate qfnames >>
        setConfig rfindfuns (List (map (\(m,n)->n++" ("++m++")") qfnames)) gp

  -- event handler for functions in find column:
  focusFunction gp =
    getValue rfindfuns gp >>= \self ->
    if null self then done else
      getFindFunctions gstate >>= \funs ->
      showModuleAndFocusFunction gp (funs!!(readNat self)) >>
      setValue rfindfuns "" gp

  -- event handler for functions in find column:
  showModuleAndFocusFunction gp (fmod,fname) =
    getContentsModule gstate >>= \cntmod ->
    if fmod == cntmod
    then getMainContents gstate >>= \(ct,cnt) ->
         let row = findFunDeclInProgText ct cnt (fmod,fname)
          in if row==0 then done else seeText rtxt (row,1) gp
    else showSource fmod gp >>
         getMainContents gstate >>= \(ct,cnt) ->
         let row = findFunDeclInProgText ct cnt (fmod,fname)
          in if row==0 then done else seeText rtxt (row,1) gp

  -- analyse all functions in the function column:
  analyzeAllFuns gp = safeIO gp $
    getSelectedModName gp >>= \mod ->
    getValue ranaall gp >>= \sela ->
    if mod==Nothing || null sela then done else
      getFuns gstate >>= \funs ->
      let (_,explanation,analysis) = allFunctionAnalyses!!readNat sela in
      setValue resultwidget explanation gp >>
      performAllAnalysis analysis (showDoing gp) (fromJust mod) funs >>= \anaresults ->
      setConfig rfun (List (map (\(prefix,func)-> prefix++" "++unqualifiedName func)
                                (zip anaresults funs))) gp >>
    setValue ranaall "" gp

  -- Perform an analysis on a module:
  performModuleAnalysis (InterfaceAnalysis ana) _ mod = do
    int <- getIntWithName gstate mod
    return (ana int)
  performModuleAnalysis (FlatCurryAnalysis ana) prt mod = do
    prog <- getProgWithName gstate prt mod
    return (ana prog)
  performModuleAnalysis (SourceCodeAnalysis ana) _ mod = do
    mbfilename <- Imports.findFileInLoadPath mod [".lcurry",".curry"]
    maybe (return (ContentsResult
                   OtherText ("Curry source file for module \""++mod++"\" not found!")))
          (\filename -> ana filename)
          mbfilename

  -- Perform an analysis to a single function declaration:
  performAnalysis (LocalAnalysis ana) prt _ fdecl = do
    prt "Analyzing..."
    return (ana fdecl)
  performAnalysis (LocalDataAnalysis ana) prt mod fdecl = do
    types <- getAllTypes gstate prt mod
    prt "Analyzing..."
    return (ana types fdecl)
  performAnalysis (GlobalAnalysis ana) prt mod fdecl = do
    funcs <- getAllFunctions gstate prt mod
    prt "Analyzing..."
    return (fromJust (lookup (funcName fdecl) (ana funcs)))
  performAnalysis (GlobalDataAnalysis ana) prt mod fdecl = do
    types <- getAllTypes gstate prt mod
    funcs <- getAllFunctions gstate prt mod
    prt "Analyzing..."
    return (fromJust (lookup (funcName fdecl) (ana types funcs)))

  -- Perform an analysis to a list of function declarations:
  performAllAnalysis (LocalAnalysis ana) prt _ fdecls = do
    prt "Analyzing..."
    return (map ana fdecls)
  performAllAnalysis (LocalDataAnalysis ana) prt mod fdecls = do
    types <- getAllTypes gstate prt mod
    prt "Analyzing..."
    return (map (ana types) fdecls)
  performAllAnalysis (GlobalAnalysis ana) prt mod fdecls =
    getAllFunctions gstate prt mod >>= \funcs ->
    prt "Analyzing..." >>
    let anaresults = ana funcs in
    return (map (\fd->fromJust (lookup (funcName fd) anaresults)) fdecls)
  performAllAnalysis (GlobalDataAnalysis ana) prt mod fdecls =
    getAllTypes gstate prt mod >>= \types ->
    getAllFunctions gstate prt mod >>= \funcs ->
    prt "Analyzing..." >>
    let anaresults = ana types funcs in
    return (map (\fd->fromJust (lookup (funcName fd) anaresults)) fdecls)


---------------------------------------------------------------------
-- find a function declaration in a program text:
findFunDeclInProgText :: ContentsKind -> String -> QName -> Int
findFunDeclInProgText CurryProg progtext fname =
  findFirstDeclLine (showCurryId (snd fname)) (lines progtext) 1
findFunDeclInProgText LCurryProg progtext fname =
  findFirstDeclLine ("> "++showCurryId (snd fname)) (lines progtext) 1
findFunDeclInProgText FlatCurryExp progtext fname =
  findFirstDeclLine ("  (Func (\""++fst fname++"\",\""++snd fname++"\")")
                    (lines progtext) 1
findFunDeclInProgText OtherText _ _ = 0

-- finds first declaration line:
findFirstDeclLine _ [] _ = 0 -- not found
findFirstDeclLine f (l:ls) n =
     if isPrefixOf f l then n else findFirstDeclLine f ls (n+1)

---------------------------------------------------------------------
-- get directory with browser sources:
getBrowserDir = do
  browserdir <- getEnviron "PAKCSHOME"
  return $ if null browserdir then "" else browserdir++"/tools/browser/"

-- order qualified names by basename first:
leqQName (mod1,n1) (mod2,n2) = leqString n1 n2 || n1==n2 && leqString mod1 mod2


---------------------------------------------------------------------
